import tensorflow as tf
from tensorflow import keras

from keras_aug.datapoints import bounding_box
from keras_aug.layers.base.vectorized_base_random_layer import (
    VectorizedBaseRandomLayer,
)
from keras_aug.utils import augmentation as augmentation_utils


@keras.utils.register_keras_serializable(package="keras_aug")
class Resize(VectorizedBaseRandomLayer):
    """Resizes the images.

    Resize will resize the images to ``(height, width)``. Set
    ``crop_to_aspect_ratio`` or ``pad_to_aspect_ratio`` to ``True`` to keep
    the aspect ratio.

    When ``crop_to_aspect_ratio`` or ``pad_to_aspect_ratio`` is set to ``True``.
    You can control the cropping position or padding position by setting
    ``position``.

    Args:
        height (int): The height of result image.
        width (int): The width of result image.
        interpolation (str, optional): The interpolation mode.
            Supported values: ``"nearest", "bilinear"``.
            Defaults to `"bilinear"`.
        antialias (bool, optional): Whether to use antialias. Defaults to
            ``False``.
        crop_to_aspect_ratio (bool, optional): If ``True``, the output images
            will be cropped to return the largest possible window in the images.
            Defaults to ``False``.
        pad_to_aspect_ratio (bool, optional): If ``True``, the output images
            will be padded to return the largest possible resize of the images.
            Defaults to ``False``.
        position (str, optional): The padding method.
            Supported values: ``"center", "top_left", "top_right", "bottom_left", "bottom_right", "random"``.
            Defaults to ``"center"``.
        padding_value (int|float, optional): The padding value.
            Defaults to ``0``.
        bounding_box_format (str, optional): The format of bounding
            boxes of input dataset. Refer
            https://github.com/james77777778/keras-aug/blob/main/keras_aug/datapoints/bounding_box/converter.py
            for more details on supported bounding box formats.
        seed (int|float, optional): The random seed. Defaults to ``None``.

    References:
        - `KerasCV <https://github.com/keras-team/keras-cv>`_
    """  # noqa: E501

    def __init__(
        self,
        height,
        width,
        interpolation="bilinear",
        antialias=False,
        crop_to_aspect_ratio=False,
        pad_to_aspect_ratio=False,
        position="center",
        padding_value=0,
        bounding_box_format=None,
        seed=None,
        **kwargs,
    ):
        super().__init__(seed=seed, **kwargs)
        if not isinstance(height, int) or not isinstance(width, int):
            raise ValueError(
                "`height` and `width` must be integer. Received: "
                f"`height`={height} `width`={width} "
            )
        self.height = height
        self.width = width
        self.interpolation = augmentation_utils.get_interpolation(interpolation)
        self.antialias = antialias
        self.position = augmentation_utils.get_padding_position(position)
        self.padding_value = padding_value
        if crop_to_aspect_ratio is True and pad_to_aspect_ratio is True:
            raise ValueError(
                "Resize expects at most one of ``crop_to_aspect_ratio`` or "
                "``pad_to_aspect_ratio`` to be ``True``"
            )
        self.crop_to_aspect_ratio = crop_to_aspect_ratio
        self.pad_to_aspect_ratio = pad_to_aspect_ratio
        self.bounding_box_format = bounding_box_format
        self.seed = seed

        # set force_output_dense_images=True because the output images must
        # have same shape (B, height, width, C)
        self.force_output_dense_images = True

    def get_random_transformation_batch(
        self, batch_size, images=None, **kwargs
    ):
        heights, widths = augmentation_utils.get_images_shape(
            images, dtype=tf.float32
        )
        # get scaled_sizes
        if self.crop_to_aspect_ratio:
            scales = tf.maximum(self.height / heights, self.width / widths)
        elif self.pad_to_aspect_ratio:
            scales = tf.minimum(self.height / heights, self.width / widths)
        else:
            return {"dummy": tf.zeros((batch_size,))}
        new_heights = tf.cast(tf.round(heights * scales), tf.int32)
        new_widths = tf.cast(tf.round(widths * scales), tf.int32)
        scaled_sizes = tf.concat((new_heights, new_widths), axis=-1)

        # get padding values
        if self.crop_to_aspect_ratio:
            tops = tf.where(
                new_heights > self.height,
                tf.cast((new_heights - self.height) / 2, tf.int32),
                0,
            )
            bottoms = tf.where(
                new_heights > self.height, new_heights - self.height - tops, 0
            )
            lefts = tf.where(
                new_widths > self.width,
                tf.cast((new_widths - self.width) / 2, tf.int32),
                0,
            )
            rights = tf.where(
                new_widths > self.width, new_widths - self.width - lefts, 0
            )
        else:
            assert self.pad_to_aspect_ratio
            tops = tf.where(
                new_heights < self.height,
                tf.cast((self.height - new_heights) / 2, tf.int32),
                0,
            )
            bottoms = tf.where(
                new_heights < self.height, self.height - new_heights - tops, 0
            )
            lefts = tf.where(
                new_widths < self.width,
                tf.cast((self.width - new_widths) / 2, tf.int32),
                0,
            )
            rights = tf.where(
                new_widths < self.width, self.width - new_widths - lefts, 0
            )
        (tops, bottoms, lefts, rights) = augmentation_utils.get_position_params(
            tops, bottoms, lefts, rights, self.position, self._random_generator
        )
        return {
            "scaled_sizes": scaled_sizes,
            "tops": tops,
            "bottoms": bottoms,
            "lefts": lefts,
            "rights": rights,
        }

    def compute_ragged_image_signature(self, images):
        return tf.RaggedTensorSpec(
            shape=(self.height, self.width, images.shape[-1]),
            ragged_rank=1,
            dtype=self.compute_dtype,
        )

    def augment_ragged_image(self, image, transformation, **kwargs):
        image = tf.expand_dims(image, axis=0)
        transformation = augmentation_utils.expand_dict_dims(
            transformation, axis=0
        )
        image = self.augment_images(
            images=image, transformations=transformation, **kwargs
        )
        return tf.squeeze(image, axis=0)

    def augment_images(self, images, transformations, **kwargs):
        # resize keeping aspect ratio
        if self.crop_to_aspect_ratio:
            images = self.resize_with_crop_to_aspect_ratio(
                images, transformations, self.interpolation
            )
        elif self.pad_to_aspect_ratio:
            images = self.resize_with_pad_to_aspect_ratio(
                images, transformations, self.interpolation, self.padding_value
            )
        else:
            # resize
            images = tf.image.resize(
                images,
                size=(self.height, self.width),
                method=self.interpolation,
                antialias=self.antialias,
            )
        images = tf.ensure_shape(
            images, shape=(None, self.height, self.width, None)
        )
        return tf.cast(images, dtype=self.compute_dtype)

    def augment_labels(self, labels, transformations, **kwargs):
        return labels

    def augment_bounding_boxes(
        self,
        bounding_boxes,
        transformations,
        images=None,
        raw_images=None,
        **kwargs,
    ):
        if self.bounding_box_format is None:
            raise ValueError(
                "`Resize()` was called with bounding boxes,"
                "but no `bounding_box_format` was specified in the constructor."
                "Please specify a bounding box format in the constructor. i.e."
                "`Resize(..., bounding_box_format='xyxy')`"
            )
        bounding_boxes = bounding_box.to_dense(bounding_boxes)

        # resize
        if not self.crop_to_aspect_ratio and not self.pad_to_aspect_ratio:
            bounding_boxes = bounding_box.convert_format(
                bounding_boxes,
                source=self.bounding_box_format,
                target="rel_xyxy",
                images=raw_images,
            )
            bounding_boxes = bounding_box.convert_format(
                bounding_boxes,
                source="rel_xyxy",
                target=self.bounding_box_format,
                dtype=self.compute_dtype,
                images=images,
            )
            return bounding_boxes

        # resize keeping aspect ratio
        bounding_boxes = bounding_box.convert_format(
            bounding_boxes,
            source=self.bounding_box_format,
            target="xyxy",
            images=raw_images,
        )
        # get new/old ratios
        ori_heights, ori_widths = augmentation_utils.get_images_shape(
            raw_images, dtype=tf.float32
        )
        scaled_sizes = transformations["scaled_sizes"]
        new_heights = tf.cast(scaled_sizes[..., 0:1], dtype=tf.float32)
        new_widths = tf.cast(scaled_sizes[..., 1:2], dtype=tf.float32)
        height_ratios = new_heights / ori_heights
        widths_ratios = new_widths / ori_widths
        tops = tf.cast(transformations["tops"], dtype=tf.float32)
        lefts = tf.cast(transformations["lefts"], dtype=tf.float32)
        # broadcast
        height_ratios = height_ratios[..., tf.newaxis]
        widths_ratios = widths_ratios[..., tf.newaxis]
        tops = tops[..., tf.newaxis]
        lefts = lefts[..., tf.newaxis]

        x1s, y1s, x2s, y2s = tf.split(bounding_boxes["boxes"], 4, axis=-1)
        if self.crop_to_aspect_ratio:
            x1s = x1s * widths_ratios - lefts
            x2s = x2s * widths_ratios - lefts
            y1s = y1s * height_ratios - tops
            y2s = y2s * height_ratios - tops
        else:
            assert self.pad_to_aspect_ratio
            x1s = x1s * widths_ratios + lefts
            x2s = x2s * widths_ratios + lefts
            y1s = y1s * height_ratios + tops
            y2s = y2s * height_ratios + tops
        boxes = tf.concat([x1s, y1s, x2s, y2s], axis=-1)
        bounding_boxes = bounding_boxes.copy()
        bounding_boxes["boxes"] = boxes
        bounding_boxes = bounding_box.clip_to_image(
            bounding_boxes,
            bounding_box_format="xyxy",
            images=images,
        )
        bounding_boxes = bounding_box.convert_format(
            bounding_boxes,
            source="xyxy",
            target=self.bounding_box_format,
            dtype=self.compute_dtype,
            images=images,
        )
        return bounding_boxes

    def compute_ragged_segmentation_mask_signature(self, segmentation_masks):
        return tf.RaggedTensorSpec(
            shape=(self.height, self.width, segmentation_masks.shape[-1]),
            ragged_rank=1,
            dtype=self.compute_dtype,
        )

    def augment_ragged_segmentation_mask(
        self, segmentation_mask, transformation, **kwargs
    ):
        segmentation_mask = tf.expand_dims(segmentation_mask, axis=0)
        transformation = augmentation_utils.expand_dict_dims(
            transformation, axis=0
        )
        segmentation_mask = self.augment_segmentation_masks(
            segmentation_masks=segmentation_mask,
            transformations=transformation,
            **kwargs,
        )
        return tf.squeeze(segmentation_mask, axis=0)

    def augment_segmentation_masks(
        self, segmentation_masks, transformations, **kwargs
    ):
        # resize keeping aspect ratio
        if self.crop_to_aspect_ratio:
            segmentation_masks = self.resize_with_crop_to_aspect_ratio(
                segmentation_masks, transformations, "nearest"
            )
        elif self.pad_to_aspect_ratio:
            segmentation_masks = self.resize_with_pad_to_aspect_ratio(
                segmentation_masks, transformations, "nearest", 0
            )
        else:
            segmentation_masks = tf.image.resize(
                segmentation_masks,
                size=(self.height, self.width),
                method="nearest",
            )
        segmentation_masks = tf.ensure_shape(
            segmentation_masks, shape=(None, self.height, self.width, None)
        )
        return tf.cast(segmentation_masks, dtype=self.compute_dtype)

    def resize_with_crop_to_aspect_ratio(self, images, transformations, method):
        batch_size = tf.shape(images)[0]
        scaled_sizes = transformations["scaled_sizes"]
        new_height = scaled_sizes[0][0]
        new_width = scaled_sizes[0][1]
        # resize
        images = tf.image.resize(
            images,
            size=(new_height, new_width),
            method=method,
            antialias=self.antialias,
        )
        # crop
        y1s = tf.cast(transformations["tops"], dtype=tf.float32)
        y2s = tf.cast(y1s + self.height, dtype=tf.float32)
        x1s = tf.cast(transformations["lefts"], dtype=tf.float32)
        x2s = tf.cast(x1s + self.width, dtype=tf.float32)
        # normalize boxes
        new_height = tf.cast(new_height, dtype=tf.float32)
        new_width = tf.cast(new_width, dtype=tf.float32)
        y1s /= new_height
        y2s /= new_height
        x1s /= new_width
        x2s /= new_width
        boxes = tf.concat([y1s, x1s, y2s, x2s], axis=-1)
        images = tf.image.crop_and_resize(
            images,
            boxes,
            tf.range(batch_size),
            [self.height, self.width],
            method="nearest",
        )
        return images

    def resize_with_pad_to_aspect_ratio(
        self, images, transformations, method, fill_value
    ):
        # resize
        scaled_sizes = transformations["scaled_sizes"]
        new_height = scaled_sizes[0][0]
        new_width = scaled_sizes[0][1]
        images = tf.image.resize(
            images,
            size=(new_height, new_width),
            method=method,
            antialias=self.antialias,
        )
        # pad
        pad_top = transformations["tops"][0][0]
        pad_bottom = transformations["bottoms"][0][0]
        pad_left = transformations["lefts"][0][0]
        pad_right = transformations["rights"][0][0]
        paddings = tf.stack(
            (
                tf.zeros(shape=(2,), dtype=tf.int32),
                tf.stack((pad_top, pad_bottom)),
                tf.stack((pad_left, pad_right)),
                tf.zeros(shape=(2,), dtype=tf.int32),
            )
        )
        images = tf.pad(images, paddings=paddings, constant_values=fill_value)
        return images

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "height": self.height,
                "width": self.width,
                "interpolation": self.interpolation,
                "antialias": self.antialias,
                "crop_to_aspect_ratio": self.crop_to_aspect_ratio,
                "pad_to_aspect_ratio": self.pad_to_aspect_ratio,
                "position": self.position,
                "padding_value": self.padding_value,
                "bounding_box_format": self.bounding_box_format,
                "seed": self.seed,
            }
        )
        return config



import math

import tensorflow as tf
from tensorflow import keras

from keras_aug.datapoints import bounding_box
from keras_aug.datapoints import image as image_utils
from keras_aug.layers.base.vectorized_base_random_layer import (
    VectorizedBaseRandomLayer,
)
from keras_aug.utils import augmentation as augmentation_utils


@keras.utils.register_keras_serializable(package="keras_aug")
class RandomAffine(VectorizedBaseRandomLayer):
    """Randomly affines transformation of the images keeping center invariant.

    Randomly affines by rotation, translation, zoom and shear. RandomAffine
    processes the images by combined transformation matrix, so it is fast.

    Args:
        rotation_factor (float|Sequence[float]|keras_aug.FactorSampler): The
            range of the degree for random rotation. When represented as a
            single float, the factor will be picked between
            ``[0.0 - lower, 0.0 + upper]``. A positive value means rotating
            counter clock-wise, while a negative value means clock-wise.
            Defaults to ``None``.
        translation_height_factor (float|Sequence[float]|keras_aug.FactorSampler): The
            range for random vertical translation. When represented as a single
            float, the factor will be picked between
            ``[0.0 - lower, 0.0 + upper]``. A negative value means shifting image
            up, while a positive value means shifting image down.
            Defaults to ``None``.
        translation_width_factor (float|Sequence[float]|keras_aug.FactorSampler): The
            range for random horizontal translation. When represented as a
            single float, the factor will be picked between
            ``[0.0 - lower, 0.0 + upper]``. A negative value means shifting
            image left, while a positive value means shifting image right.
            Defaults to ``None``.
        zoom_height_factor (float|Sequence[float]|keras_aug.FactorSampler): The
            range for random vertical zoom. When represented as a
            single float, the factor will be picked between
            ``[1.0 - lower, 1.0 + upper]``. A negative value means zooming in
            while a positive value means zooming out. Defaults to ``None``.
        zoom_width_factor (float|Sequence[float]|keras_aug.FactorSampler): The
            range for random horizontal zoom. When represented as a
            single float, the factor will be picked between
            ``[1.0 - lower, 1.0 + upper]``. A negative value means zooming in
            while a positive value means zooming out. Defaults to ``None``.
        shear_height_factor (float|Sequence[float]|keras_aug.FactorSampler): The
            range for random vertical shear. When represented as a
            single float, the factor will be picked between
            ``[0.0 - lower, 0.0 + upper]``. Defaults to ``None``.
        shear_width_factor (float|Sequence[float]|keras_aug.FactorSampler): The
            range for random horizontal shear. When represented as a
            single float, the factor will be picked between
            ``[0.0 - lower, 0.0 + upper]``. Defaults to ``None``.
        same_zoom_factor (bool, optional): If True, the zoom factor sampled from
            ``zoom_height_factor`` will be applied to both height and width.
            It is useful to keep aspect ratio. Defaults to ``False``.
        interpolation (str, optional): The interpolation mode. Supported values:
            ``"nearest", "bilinear"``. Defaults to `"bilinear"`.
        fill_mode (str, optional): The fill mode. Supported values:
            ``"constant", "reflect", "wrap", "nearest"``. Defaults to
            ``"constant"``.
        fill_value (int|float, optional): The value to be filled outside the
            boundaries when ``fill_mode="constant"``. Defaults to ``0``.
        bounding_box_format (str, optional): The format of bounding
            boxes of input dataset. Refer
            https://github.com/james77777778/keras-aug/blob/main/keras_aug/datapoints/bounding_box/converter.py
            for more details on supported bounding box formats.
        bounding_box_min_area_ratio (float, optional): The threshold to
            apply sanitize_bounding_boxes. Defaults to ``None``.
        bounding_box_max_aspect_ratio (float, optional): The threshold to
            apply sanitize_bounding_boxes. Defaults to ``None``.
        seed (int|float, optional): The random seed. Defaults to ``None``.

    References:
        - `KerasCV <https://github.com/keras-team/keras-cv>`_
    """  # noqa: E501

    def __init__(
        self,
        rotation_factor=None,
        translation_height_factor=None,
        translation_width_factor=None,
        zoom_height_factor=None,
        zoom_width_factor=None,
        shear_height_factor=None,
        shear_width_factor=None,
        same_zoom_factor=False,
        interpolation="bilinear",
        fill_mode="constant",
        fill_value=0,
        bounding_box_format=None,
        bounding_box_min_area_ratio=None,
        bounding_box_max_aspect_ratio=None,
        seed=None,
        **kwargs,
    ):
        super().__init__(seed=seed, **kwargs)
        self.rotation_factor = None
        self.translation_height_factor = None
        self.translation_width_factor = None
        self.zoom_height_factor = None
        self.zoom_width_factor = None
        self.shear_height_factor = None
        self.shear_width_factor = None
        # rotation
        if rotation_factor is not None:
            self.rotation_factor = augmentation_utils.parse_factor(
                rotation_factor,
                min_value=-180,
                max_value=180,
                center_value=0,
                seed=seed,
            )
        # translation
        if translation_height_factor is not None:
            self.translation_height_factor = augmentation_utils.parse_factor(
                translation_height_factor,
                min_value=-1,
                max_value=1,
                center_value=0.0,
                seed=seed,
            )
        if translation_width_factor is not None:
            self.translation_width_factor = augmentation_utils.parse_factor(
                translation_width_factor,
                min_value=-1,
                max_value=1,
                center_value=0.0,
                seed=seed,
            )
        # zoom
        if zoom_height_factor is not None:
            self.zoom_height_factor = augmentation_utils.parse_factor(
                zoom_height_factor,
                min_value=0,
                max_value=None,
                center_value=1.0,
                seed=seed,
            )
        if zoom_width_factor is not None:
            self.zoom_width_factor = augmentation_utils.parse_factor(
                zoom_width_factor,
                min_value=0,
                max_value=None,
                center_value=1.0,
                seed=seed,
            )
        # shear
        if shear_height_factor is not None:
            self.shear_height_factor = augmentation_utils.parse_factor(
                shear_height_factor,
                min_value=-1,
                max_value=1,
                center_value=0.0,
                seed=seed,
            )
        if shear_width_factor is not None:
            self.shear_width_factor = augmentation_utils.parse_factor(
                shear_width_factor,
                min_value=-1,
                max_value=1,
                center_value=0.0,
                seed=seed,
            )

        self.same_zoom_factor = same_zoom_factor

        augmentation_utils.check_fill_mode_and_interpolation(
            fill_mode, interpolation
        )
        self.interpolation = interpolation
        self.fill_mode = fill_mode
        self.fill_value = fill_value
        self.bounding_box_format = bounding_box_format
        self.bounding_box_min_area_ratio = bounding_box_min_area_ratio
        self.bounding_box_max_aspect_ratio = bounding_box_max_aspect_ratio
        self.seed = seed

        # decide whether to enable the augmentation
        self._enable_rotation = augmentation_utils.is_factor_working(
            self.rotation_factor, not_working_value=0.0
        )
        _enable_translation_height = augmentation_utils.is_factor_working(
            self.translation_height_factor, not_working_value=0.0
        )
        _enable_translation_width = augmentation_utils.is_factor_working(
            self.translation_width_factor, not_working_value=0.0
        )
        self._enable_translation = (
            _enable_translation_height or _enable_translation_width
        )
        _enable_zoom_height = augmentation_utils.is_factor_working(
            self.zoom_height_factor, not_working_value=0.0
        )
        _enable_zoom_width = augmentation_utils.is_factor_working(
            self.zoom_width_factor, not_working_value=0.0
        )
        self._enable_zoom = _enable_zoom_height or _enable_zoom_width
        _enable_shear_height = augmentation_utils.is_factor_working(
            self.shear_height_factor, not_working_value=0.0
        )
        _enable_shear_width = augmentation_utils.is_factor_working(
            self.shear_width_factor, not_working_value=0.0
        )
        self._enable_shear = _enable_shear_height or _enable_shear_width

    def get_random_transformation_batch(
        self, batch_size, images=None, **kwargs
    ):
        # cast to float32 to avoid numerical issue
        heights, widths = augmentation_utils.get_images_shape(
            images, dtype=tf.float32
        )
        factor_shape = (batch_size, 1)
        # dummy
        angles = tf.zeros(factor_shape, dtype=tf.float32)
        translation_heights = tf.zeros(factor_shape, dtype=tf.float32)
        translation_widths = tf.zeros(factor_shape, dtype=tf.float32)
        zoom_heights = tf.zeros(factor_shape, dtype=tf.float32)
        zoom_widths = tf.zeros(factor_shape, dtype=tf.float32)
        shear_heights = tf.zeros(factor_shape, dtype=tf.float32)
        shear_widths = tf.zeros(factor_shape, dtype=tf.float32)

        if self._enable_rotation:
            angles = self.rotation_factor(factor_shape, dtype=tf.float32)
        if self._enable_translation:
            translation_heights = self.translation_height_factor(
                factor_shape, dtype=tf.float32
            )
            translation_widths = self.translation_width_factor(
                factor_shape, dtype=tf.float32
            )
        if self._enable_zoom:
            zoom_heights = self.zoom_height_factor(
                factor_shape, dtype=tf.float32
            )
            if self.same_zoom_factor:
                zoom_widths = zoom_heights
            else:
                zoom_widths = self.zoom_width_factor(
                    factor_shape, dtype=tf.float32
                )
        if self._enable_shear:
            shear_heights = self.shear_height_factor(
                factor_shape, dtype=tf.float32
            )
            shear_widths = self.shear_width_factor(
                factor_shape, dtype=tf.float32
            )

        angles = angles / 360.0 * 2.0 * math.pi
        translations = tf.concat(
            [translation_widths, translation_heights], axis=1
        )
        zooms = tf.concat([zoom_widths, zoom_heights], axis=1)
        shears = tf.concat([shear_widths, shear_heights], axis=1)

        # start from identity matrixes:
        #     [[1 0 0]
        #      [0 1 0]
        #      [0 0 1]]
        identity_matrixes = tf.concat(
            [
                tf.ones((batch_size, 1), dtype=tf.float32),
                tf.zeros((batch_size, 3), dtype=tf.float32),
                tf.ones((batch_size, 1), dtype=tf.float32),
                tf.zeros((batch_size, 3), dtype=tf.float32),
                tf.ones((batch_size, 1), dtype=tf.float32),
            ],
            axis=1,
        )
        combined_matrixes = tf.reshape(identity_matrixes, (batch_size, 3, 3))
        # process zoom
        if self._enable_zoom:
            zoom_matrixes = image_utils.get_zoom_matrix(
                zooms, heights, widths, to_square=True
            )
            combined_matrixes = zoom_matrixes @ combined_matrixes
        # process rotations
        if self._enable_rotation:
            rotation_matrixes = image_utils.get_rotation_matrix(
                angles, heights, widths, to_square=True
            )
            combined_matrixes = rotation_matrixes @ combined_matrixes
        # process shear
        if self._enable_shear:
            shear_matrixes = image_utils.get_shear_matrix(
                shears, to_square=True
            )
            combined_matrixes = shear_matrixes @ combined_matrixes
        # process translations
        if self._enable_translation:
            translation_matrixes = image_utils.get_translation_matrix(
                translations, heights, widths, to_square=True
            )
            combined_matrixes = translation_matrixes @ combined_matrixes
        return {
            "angles": angles,
            "translations": translations,
            "zooms": zooms,
            "shears": shears,
            "combined_matrixes": combined_matrixes,  # (batch_size, 3, 3)
        }

    def augment_ragged_image(self, image, transformation, **kwargs):
        image = tf.expand_dims(image, axis=0)
        transformation = augmentation_utils.expand_dict_dims(
            transformation, axis=0
        )
        image = self.augment_images(
            images=image, transformations=transformation, **kwargs
        )
        return tf.squeeze(image, axis=0)

    def augment_images(self, images, transformations, **kwargs):
        ori_shape = images.shape
        batch_size = tf.shape(images)[0]
        combined_matrixes = transformations["combined_matrixes"]
        combined_matrixes = tf.reshape(
            combined_matrixes, shape=(batch_size, -1)
        )
        combined_matrixes = combined_matrixes[:, :-1]

        # tf.raw_ops.ImageProjectiveTransformV3 not support bfloat16
        if images.dtype == tf.bfloat16:
            images = tf.cast(images, dtype=tf.float32)
        images = image_utils.transform(
            images,
            combined_matrixes,
            fill_mode=self.fill_mode,
            fill_value=self.fill_value,
            interpolation=self.interpolation,
        )
        images = tf.ensure_shape(images, shape=ori_shape)
        return tf.cast(images, dtype=self.compute_dtype)

    def augment_labels(self, labels, transformations, **kwargs):
        return labels

    def augment_bounding_boxes(
        self, bounding_boxes, transformations, raw_images=None, **kwargs
    ):
        if self.bounding_box_format is None:
            raise ValueError(
                "`RandomAffine()` was called with bounding boxes,"
                "but no `bounding_box_format` was specified in the constructor."
                "Please specify a bounding box format in the constructor. i.e."
                "`RandomAffine(bounding_box_format='xyxy')`"
            )
        # cast to float32 to avoid numerical issue
        heights, widths = augmentation_utils.get_images_shape(
            raw_images, dtype=tf.float32
        )
        bounding_boxes = bounding_box.to_dense(bounding_boxes)
        bounding_boxes = bounding_box.convert_format(
            bounding_boxes,
            source=self.bounding_box_format,
            target="xyxy",
            images=raw_images,
            dtype=tf.float32,
        )
        boxes = bounding_boxes["boxes"]
        original_bounding_boxes = bounding_boxes.copy()
        # process rotations
        if self._enable_rotation:
            origin_x = widths / 2
            origin_y = heights / 2
            angles = -transformations["angles"]
            angles = angles[:, tf.newaxis, tf.newaxis]
            # points: (batch_size, max_num_boxes, 4, 2)
            points = tf.stack(
                [
                    tf.stack([boxes[..., 0], boxes[..., 1]], axis=2),
                    tf.stack([boxes[..., 2], boxes[..., 1]], axis=2),
                    tf.stack([boxes[..., 2], boxes[..., 3]], axis=2),
                    tf.stack([boxes[..., 0], boxes[..., 3]], axis=2),
                ],
                axis=2,
            )
            point_x_offsets = points[..., 0] - origin_x[..., tf.newaxis]
            point_y_offsets = points[..., 1] - origin_y[..., tf.newaxis]
            new_x = (
                origin_x[..., tf.newaxis, tf.newaxis]
                + tf.multiply(tf.cos(angles), point_x_offsets[..., tf.newaxis])
                - tf.multiply(tf.sin(angles), point_y_offsets[..., tf.newaxis])
            )
            new_y = (
                origin_y[..., tf.newaxis, tf.newaxis]
                + tf.multiply(tf.sin(angles), point_x_offsets[..., tf.newaxis])
                + tf.multiply(tf.cos(angles), point_y_offsets[..., tf.newaxis])
            )
            out = tf.concat([new_x, new_y], axis=3)
            min_cordinates = tf.math.reduce_min(out, axis=2)
            max_cordinates = tf.math.reduce_max(out, axis=2)
            boxes = tf.concat([min_cordinates, max_cordinates], axis=2)
        # process translations
        if self._enable_translation:
            translations = transformations["translations"]
            translation_widths = tf.expand_dims(
                translations[:, 0:1] * widths, axis=-1
            )
            translation_heights = tf.expand_dims(
                translations[:, 1:2] * heights, axis=-1
            )
            x1s, y1s, x2s, y2s = tf.split(boxes, 4, axis=-1)
            x1s = x1s + translation_widths
            y1s = y1s + translation_heights
            x2s = x2s + translation_widths
            y2s = y2s + translation_heights
            boxes = tf.concat([x1s, y1s, x2s, y2s], axis=-1)
        # process shear
        if self._enable_shear:
            shears = transformations["shears"]
            shear_widths = tf.expand_dims(shears[:, 0:1], axis=-1)
            shear_heights = tf.expand_dims(shears[:, 1:2], axis=-1)
            _x1s, _y1s, _x2s, _y2s = tf.split(boxes, 4, axis=-1)
            # x1, x2
            x1_tops = _x1s - (shear_widths * _y1s)
            x1_bottoms = _x1s - (shear_widths * _y2s)
            x1s = tf.where(shear_widths < 0, x1_tops, x1_bottoms)
            x2_tops = _x2s - (shear_widths * _y1s)
            x2_bottoms = _x2s - (shear_widths * _y2s)
            x2s = tf.where(shear_widths < 0, x2_bottoms, x2_tops)
            # y1, y2
            y1_lefts = _y1s - (shear_heights * _x1s)
            y1_rights = _y1s - (shear_heights * _x2s)
            y1s = tf.where(shear_heights > 0, y1_rights, y1_lefts)
            y2_lefts = _y2s - (shear_heights * _x1s)
            y2_rights = _y2s - (shear_heights * _x2s)
            y2s = tf.where(shear_heights > 0, y2_lefts, y2_rights)
            boxes = tf.concat([x1s, y1s, x2s, y2s], axis=-1)
        # process zoom
        if self._enable_zoom:
            zooms = transformations["zooms"]
            zoom_widths = tf.expand_dims(zooms[:, 0:1], axis=-1)
            zoom_heights = tf.expand_dims(zooms[:, 1:2], axis=-1)
            x1s, y1s, x2s, y2s = tf.split(boxes, 4, axis=-1)
            x_offsets = ((tf.expand_dims(widths, axis=-1) - 1.0) / 2.0) * (
                1.0 - zoom_widths
            )
            y_offsets = ((tf.expand_dims(heights, axis=-1) - 1.0) / 2.0) * (
                1.0 - zoom_heights
            )
            x1s = (x1s - x_offsets) / zoom_widths
            x2s = (x2s - x_offsets) / zoom_widths
            y1s = (y1s - y_offsets) / zoom_heights
            y2s = (y2s - y_offsets) / zoom_heights
            boxes = tf.concat([x1s, y1s, x2s, y2s], axis=-1)

        bounding_boxes = bounding_boxes.copy()
        bounding_boxes["boxes"] = boxes
        bounding_boxes = bounding_box.clip_to_image(
            bounding_boxes,
            bounding_box_format="xyxy",
            images=raw_images,
        )
        bounding_boxes = bounding_box.sanitize_bounding_boxes(
            bounding_boxes,
            min_area_ratio=self.bounding_box_min_area_ratio,
            max_aspect_ratio=self.bounding_box_max_aspect_ratio,
            bounding_box_format="xyxy",
            reference_bounding_boxes=original_bounding_boxes,
            images=raw_images,
            reference_images=raw_images,
        )
        bounding_boxes = bounding_box.convert_format(
            bounding_boxes,
            source="xyxy",
            target=self.bounding_box_format,
            dtype=self.compute_dtype,
            images=raw_images,
        )
        return bounding_boxes

    def augment_ragged_segmentation_mask(
        self, segmentation_mask, transformation, **kwargs
    ):
        segmentation_mask = tf.expand_dims(segmentation_mask, axis=0)
        transformation = augmentation_utils.expand_dict_dims(
            transformation, axis=0
        )
        segmentation_mask = self.augment_segmentation_masks(
            segmentation_masks=segmentation_mask,
            transformations=transformation,
            **kwargs,
        )
        return tf.squeeze(segmentation_mask, axis=0)

    def augment_segmentation_masks(
        self, segmentation_masks, transformations, **kwargs
    ):
        original_shape = segmentation_masks.shape
        batch_size = tf.shape(segmentation_masks)[0]
        combined_matrixes = transformations["combined_matrixes"]
        combined_matrixes = tf.reshape(
            combined_matrixes, shape=(batch_size, -1)
        )
        combined_matrixes = combined_matrixes[:, :-1]

        # tf.raw_ops.ImageProjectiveTransformV3 not support bfloat16
        if segmentation_masks.dtype == tf.bfloat16:
            segmentation_masks = tf.cast(segmentation_masks, dtype=tf.float32)
        segmentation_masks = image_utils.transform(
            segmentation_masks,
            combined_matrixes,
            fill_mode=self.fill_mode,
            fill_value=0,
            interpolation="nearest",
        )
        segmentation_masks = tf.ensure_shape(
            segmentation_masks, shape=original_shape
        )
        return tf.cast(segmentation_masks, dtype=self.compute_dtype)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "rotation_factor": self.rotation_factor,
                "translation_height_factor": self.translation_height_factor,
                "translation_width_factor": self.translation_width_factor,
                "zoom_height_factor": self.zoom_height_factor,
                "zoom_width_factor": self.zoom_width_factor,
                "shear_height_factor": self.shear_height_factor,
                "shear_width_factor": self.shear_width_factor,
                "same_zoom_factor": self.same_zoom_factor,
                "fill_mode": self.fill_mode,
                "fill_value": self.fill_value,
                "interpolation": self.interpolation,
                "bounding_box_format": self.bounding_box_format,
                "bounding_box_min_area_ratio": self.bounding_box_min_area_ratio,  # noqa: E501
                "bounding_box_max_aspect_ratio": self.bounding_box_max_aspect_ratio,  # noqa: E501
                "seed": self.seed,
            }
        )
        return config



import tensorflow as tf

from keras_aug import layers


class NormalizeTest(tf.test.TestCase):
    regular_args = {
        "value_range": (0, 255),
        "mean": (0.485, 0.456, 0.406),
        "std": (0.229, 0.224, 0.225),
    }
    no_aug_args = {
        "value_range": (0, 1),
        "mean": (0.0, 0.0, 0.0),
        "std": (1.0, 1.0, 1.0),
    }

    def test_no_adjustment(self):
        image_shape = (4, 8, 8, 3)
        image = tf.random.uniform(shape=image_shape)

        layer = layers.Normalize(**self.no_aug_args)
        output = layer(image)

        self.assertAllClose(image, output)

    def test_adjustment_for_non_rgb_value_range(self):
        image_shape = (4, 8, 8, 3)
        # Value range (0, 100)
        image = tf.random.uniform(shape=image_shape) * 100.0

        layer = layers.Normalize(**self.regular_args)
        output = layer(image)
        self.assertNotAllClose(image, output)

    def test_normalize_output(self):
        image_shape = (4, 8, 8, 3)
        image = tf.random.uniform(shape=image_shape) * 255.0
        mean = tf.convert_to_tensor(self.regular_args["mean"])
        mean = tf.reshape(mean, shape=(1, 1, 1, 3))
        std = tf.convert_to_tensor(self.regular_args["std"])
        std = tf.reshape(std, shape=(1, 1, 1, 3))
        expected_output = (image - mean * 255.0) / (std * 255.0)

        layer = layers.Normalize(**self.regular_args)
        output = layer(image)

        self.assertAllClose(output, expected_output)



from functools import partial

import tensorflow as tf

from keras_aug import layers

"""
equalize is borrowed from:
https://github.com/tensorflow/addons/blob/v0.17.0/tensorflow_addons/image/color_ops.py#L62-L84
"""  # noqa: E501


def _scale_channel(image, channel: int, bins: int = 256) -> tf.Tensor:
    """Scale the data in the channel to implement equalize."""
    image_dtype = image.dtype
    image = tf.cast(image[:, :, channel], tf.int32)

    # Compute the histogram of the image channel.
    histo = tf.histogram_fixed_width(image, [0, bins - 1], nbins=bins)

    # For the purposes of computing the step, filter out the nonzeros.
    nonzero_histo = tf.boolean_mask(histo, histo != 0)
    step = (tf.reduce_sum(nonzero_histo) - nonzero_histo[-1]) // (bins - 1)

    # If step is zero, return the original image.  Otherwise, build
    # lut from the full histogram and step and then index from it.
    if step == 0:
        result = image
    else:
        lut_values = (tf.cumsum(histo, exclusive=True) + (step // 2)) // step
        lut_values = tf.clip_by_value(lut_values, 0, bins - 1)
        result = tf.gather(lut_values, image)

    return tf.cast(result, image_dtype)


def _equalize_image(image, bins: int = 256) -> tf.Tensor:
    """Implements Equalize function from PIL using TF ops."""
    image = tf.stack(
        [_scale_channel(image, c, bins) for c in range(image.shape[-1])], -1
    )
    return image


def equalize(image, bins: int = 256) -> tf.Tensor:
    """Equalize image(s)
    Args:
      images: A tensor of shape
          `(num_images, num_rows, num_columns, num_channels)` (NHWC), or
          `(num_rows, num_columns, num_channels)` (HWC), or
          `(num_rows, num_columns)` (HW). The rank must be statically known (the
          shape is not `TensorShape(None)`).
      bins: The number of bins in the histogram.
      name: The name of the op.
    Returns:
      Image(s) with the same type and shape as `images`, equalized.
    """
    fn = partial(_equalize_image)
    image = tf.map_fn(lambda x: fn(x, bins), image)
    return image


class RandomCLAHETest(tf.test.TestCase):
    regular_args = {
        "value_range": (0, 255),
        "factor": (2, 4),
        "tile_grid_size": (8, 8),
    }

    def test_adjustment_for_non_rgb_value_range(self):
        image_shape = (4, 8, 8, 3)
        # Value range (0, 100)
        image = tf.random.uniform(shape=image_shape) * 100.0
        args = self.regular_args.copy()
        args.update({"value_range": (0, 100)})
        layer = layers.RandomCLAHE(**args)

        output = layer(image)

        self.assertNotAllClose(image, output)

    def test_random_clahe_output(self):
        image_shape = (4, 8, 8, 3)
        image = tf.random.uniform(shape=image_shape) * 255.0
        image = tf.cast(image, dtype=tf.uint8)
        args = self.regular_args.copy()
        args.update({"factor": (1, 1), "tile_grid_size": (1, 1)})
        layer = layers.RandomCLAHE(**args)
        equalized = equalize(image)

        output = layer(image)

        self.assertAllClose(output, equalized, atol=1)



try:
    import tensorflow_datasets
except ImportError:
    tensorflow_datasets = None
try:
    import keras_cv
except ImportError:
    keras_cv = None


def assert_tfds_installed(symbol_name):
    if tensorflow_datasets is None:
        raise ImportError(
            f"{symbol_name} requires the `tensorflow_datasets` package. "
            "Please install the package using "
            "`pip install tensorflow_datasets`."
        )


def assert_kerascv_installed(symbol_name):
    if keras_cv is None:
        raise ImportError(
            f"{symbol_name} requires the `keras_cv` package. "
            "Please install the package using `pip install keras_cv`."
        )



import tensorflow as tf
from absl.testing import parameterized

from keras_aug.datapoints import image


class ImageOpsTest(tf.test.TestCase, parameterized.TestCase):
    def test_blend(self):
        ones = tf.ones(shape=(2, 4, 4, 3))
        twos = tf.ones(shape=(2, 4, 4, 3)) * 2
        ratios = tf.ones(shape=(2, 1, 1, 1)) * 0.3
        expected_result = ratios * ones + (1.0 - ratios) * twos

        result = image.blend(twos, ones, ratios)

        self.assertAllEqual(result, expected_result)



import keras

from keras_aug.core.factor_sampler.factor_sampler import FactorSampler
from keras_aug.core.keras_random_generator import KerasRandomGenerator


@keras.utils.register_keras_serializable(package="keras_aug")
class UniformFactorSampler(FactorSampler):
    """UniformFactorSampler samples factors uniformly from a range.

    Args:
        lower (float): the lower bound of values returned from ``__call__()``.
        upper (float): the upper bound of values returned from ``__call__()``.
        seed (int|float, optional): The random seed. Defaults to ``None``.

    References:
        - `KerasCV <https://github.com/keras-team/keras-cv>`_
    """

    def __init__(self, lower, upper, seed=None):
        self.lower = lower
        self.upper = upper
        self.seed = seed
        self.rng = KerasRandomGenerator(seed=seed, rng_type="stateless")

    def __call__(self, shape=(), dtype="float32"):
        return self.rng.random_uniform(
            shape,
            minval=self.lower,
            maxval=self.upper,
            dtype=dtype,
        )

    def get_config(self):
        return {
            "lower": self.lower,
            "upper": self.upper,
            "seed": self.seed,
        }



import copy
import inspect

import tensorflow as tf
from absl.testing import parameterized

from keras_aug import layers
from keras_aug.layers import augmentation
from keras_aug.layers import preprocessing
from keras_aug.utils.augmentation import BOUNDING_BOXES
from keras_aug.utils.augmentation import IMAGES
from keras_aug.utils.augmentation import LABELS

SEED = 2025
#   (
#       name,
#       layer_cls,
#       args,
#       is_bbox_compatible,
#   )
# all configurations should be expanded for readability
GENERAL_TESTS = [
    (
        "AugMix",
        layers.AugMix,
        {"value_range": (0, 255)},
        False,
    ),
    (
        "RandAugment",
        layers.RandAugment,
        {"value_range": (0, 255)},
        True,
    ),
    (
        "TrivialAugmentWide",
        layers.TrivialAugmentWide,
        {"value_range": (0, 255)},
        True,
    ),
    (
        "RandomAffine",
        layers.RandomAffine,
        {
            "rotation_factor": 10,
            "translation_height_factor": 0.1,
            "translation_width_factor": 0.1,
            "zoom_height_factor": 0.1,
            "zoom_width_factor": 0.1,
            "shear_height_factor": 0.1,
            "shear_width_factor": 0.1,
        },
        True,
    ),
    (
        "RandomCrop",
        layers.RandomCrop,
        {"height": 2, "width": 2},
        True,
    ),
    (
        "RandomCropAndResize",
        layers.RandomCropAndResize,
        {
            "height": 2,
            "width": 2,
            "crop_area_factor": (0.8, 1.0),
            "aspect_ratio_factor": (3 / 4, 4 / 3),
        },
        True,
    ),
    (
        "RandomFlip",
        layers.RandomFlip,
        {"mode": "horizontal_and_vertical"},
        True,
    ),
    (
        "RandomResize",
        layers.RandomResize,
        {"heights": [2]},
        True,
    ),
    (
        "RandomRotate",
        layers.RandomRotate,
        {"factor": 10},
        True,
    ),
    (
        "RandomZoomAndCrop",
        layers.RandomZoomAndCrop,
        {"height": 2, "width": 2, "scale_factor": (0.8, 1.25)},
        True,
    ),
    (
        "ChannelShuffle",
        layers.ChannelShuffle,
        {"groups": 3},
        True,
    ),
    (
        "RandomBlur",
        layers.RandomBlur,
        {"factor": (3, 7), "seed": 2024},
        True,
    ),
    (
        "RandomChannelShift",
        layers.RandomChannelShift,
        {"value_range": (0, 255), "factor": 0.1},
        True,
    ),
    (
        "RandomCLAHE",
        layers.RandomCLAHE,
        {
            "value_range": (0, 255),
            "factor": (1, 100),
            "tile_grid_size": (4, 4),
            "seed": 2024,  # manually set
        },
        True,
    ),
    (
        "RandomColorJitter",
        layers.RandomColorJitter,
        {
            "value_range": (0, 255),
            "brightness_factor": 0.1,
            "contrast_factor": 0.1,
            "saturation_factor": 0.1,
            "hue_factor": 0.1,
        },
        True,
    ),
    (
        "RandomGamma",
        layers.RandomGamma,
        {"value_range": (0, 255), "factor": 0.1},
        True,
    ),
    (
        "RandomGaussianBlur",
        layers.RandomGaussianBlur,
        {"kernel_size": 3, "factor": 2.0},
        True,
    ),
    (
        "RandomHSV",
        layers.RandomHSV,
        {
            "value_range": (0, 255),
            "hue_factor": 0.1,
            "saturation_factor": 0.1,
            "value_factor": 0.1,
        },
        True,
    ),
    (
        "RandomJpegQuality",
        layers.RandomJpegQuality,
        {
            "value_range": (0, 255),
            "factor": (75, 100),
        },
        True,
    ),
    (
        "RandomPosterize",
        layers.RandomPosterize,
        {"value_range": (0, 255), "factor": (5, 8)},
        True,
    ),
    (
        "RandomSharpness",
        layers.RandomSharpness,
        {"value_range": (0, 255), "factor": 0.1},
        True,
    ),
    (
        "RandomSolarize",
        layers.RandomSolarize,
        {
            "value_range": (0, 255),
            "threshold_factor": 10,
            "addition_factor": 10,
        },
        True,
    ),
    (
        "CutMix",
        layers.CutMix,
        {"alpha": 1.0},
        False,
    ),
    (
        "MixUp",
        layers.MixUp,
        {},
        True,
    ),
    (
        "Mosaic",
        layers.Mosaic,
        {
            "height": 100,
            "width": 100,
        },
        True,
    ),
    (
        "RandomChannelDropout",
        layers.RandomChannelDropout,
        {"seed": 2024},
        True,
    ),
    (
        "RandomCutout",
        layers.RandomCutout,
        {"height_factor": 0.3, "width_factor": 0.3},
        True,
    ),
    (
        "RandomErase",
        layers.RandomErase,
        {"area_factor": (0.02, 0.4), "aspect_ratio_factor": (0.3, 1.0 / 0.3)},
        False,
    ),
    (
        "RandomGridMask",
        layers.RandomGridMask,
        {
            "size_factor": (0.5, 1.0),
            "ratio_factor": (0.6, 0.6),
            "rotation_factor": (-10, 10),
        },
        True,
    ),
    (
        "RandomApply",
        layers.RandomApply,
        {
            "layer": layers.RandomColorJitter(
                value_range=(0, 255), brightness_factor=(0.5, 1.5), seed=SEED
            ),
            "seed": 2024,
        },
        True,
    ),
    (
        "RandomChoice",
        layers.RandomChoice,
        {
            "layers": [
                layers.RandomColorJitter(
                    value_range=(0, 255), brightness_factor=(0.5, 0.5)
                ),
                layers.RandomColorJitter(
                    value_range=(0, 255), brightness_factor=(1.5, 1.5)
                ),
            ],
            "seed": 2024,
        },
        True,
    ),
    (
        "RepeatedAugment",
        layers.RepeatedAugment,
        {
            "layers": [
                layers.RandomColorJitter(
                    value_range=(0, 255), brightness_factor=(1.5, 1.5)
                ),
                layers.RandomColorJitter(
                    value_range=(0, 255), contrast_factor=(1.5, 1.5)
                ),
            ]
        },
        True,
    ),
    (
        "CenterCrop",
        layers.CenterCrop,
        {"height": 2, "width": 2},
        True,
    ),
    (
        "PadIfNeeded",
        layers.PadIfNeeded,
        {"min_height": 2, "min_width": 2},
        True,
    ),
    (
        "Resize",
        layers.Resize,
        {"height": 2, "width": 2},
        True,
    ),
    (
        "AutoContrast",
        layers.AutoContrast,
        {"value_range": (0, 255)},
        True,
    ),
    (
        "Equalize",
        layers.Equalize,
        {"value_range": (0, 255)},
        True,
    ),
    (
        "Grayscale",
        layers.Grayscale,
        {"output_channels": 3},
        True,
    ),
    (
        "Identity",
        layers.Identity,
        {},
        True,
    ),
    (
        "Invert",
        layers.Invert,
        {"value_range": (0, 255)},
        True,
    ),
    (
        "Normalize",
        layers.Normalize,
        {"value_range": (0, 255)},
        True,
    ),
    (
        "Rescale",
        layers.Rescale,
        {"scale": 1.0 / 255.0},
        True,
    ),
    (
        "SanitizeBoundingBox",
        layers.SanitizeBoundingBox,
        {"min_size": 10},
        True,
    ),
]

NO_PRESERVED_SHAPE = [
    layers.RandomCrop,
    layers.RandomCropAndResize,
    layers.RandomResize,
    layers.RandomZoomAndCrop,
    layers.Mosaic,
    layers.RepeatedAugment,
    layers.CenterCrop,
    layers.Resize,
]

NO_UINT8 = [
    layers.AugMix,  # alpha
    layers.RandAugment,  # stateless_random_uniform
    layers.TrivialAugmentWide,  # stateless_random_uniform
    layers.RandomSharpness,  # stateless_random_uniform
    layers.RandomSolarize,  # stateless_random_uniform
    layers.CutMix,  # tf.convert_to_tensor
    layers.MixUp,  # tf.convert_to_tensor
    layers.RandomCutout,  # tf.where with -1 (invalid bbox)
    layers.RandomHSV,  # stateless_random_uniform
    layers.RandomChannelShift,  # stateless_random_uniform
    layers.RandomColorJitter,  # stateless_random_uniform
    layers.RandomGamma,  # stateless_random_uniform
    layers.RandomGaussianBlur,  # tf.nn.depthwise_conv2d
    layers.RandomJpegQuality,  # image_utils.transform_value_range
    layers.AutoContrast,  # tf.convert_to_tensor
    layers.Grayscale,  # tf.mul
    layers.Normalize,  # mean, std
]

SKIP_DTYPE = [
    # it is hard to change dtype in runtime
    layers.RandomApply,
    layers.RepeatedAugment,
]

ALWAYS_SAME_OUTPUT_WITHIN_BATCH_LAYERS = [
    layers.RandomResize,  # same size in the batch
    layers.CutMix,  # cannot test CutMix with same images
    layers.MixUp,  # cannot test MixUp with same images
    layers.CenterCrop,
    layers.PadIfNeeded,
    layers.Resize,
    layers.AutoContrast,
    layers.Equalize,
    layers.Grayscale,
    layers.Identity,
    layers.Invert,
    layers.Normalize,
    layers.Rescale,
    layers.SanitizeBoundingBox,
]


class OutputCommonTest(tf.test.TestCase, parameterized.TestCase):
    def test_all_2d_aug_layers_included(self):
        base_cls = layers.VectorizedBaseRandomLayer
        cls_spaces = [augmentation, preprocessing]
        all_2d_aug_layers = []
        for cls_space in cls_spaces:
            all_2d_aug_layers.extend(
                inspect.getmembers(cls_space, predicate=inspect.isclass)
            )
        all_2d_aug_layer_names = set(
            item[0]
            for item in all_2d_aug_layers
            if issubclass(item[1], base_cls)
        )

        general_names = set(item[0] for item in GENERAL_TESTS)
        all_test_names = general_names

        for name in all_2d_aug_layer_names:
            self.assertIn(name, all_test_names, msg=f"{name} not found")

    @parameterized.named_parameters(*GENERAL_TESTS)
    def test_preserves_output_shape(self, layer_cls, args, is_bbox_compatible):
        images = tf.random.uniform(shape=(2, 16, 16, 3), seed=SEED) * 255.0
        labels = tf.random.uniform(shape=(2, 1), seed=SEED) * 10.0
        bounding_boxes = {
            "boxes": tf.ragged.constant(
                [
                    [[0, 0, 2, 2], [0, 0, 16, 16]],
                    [[2, 5, 1, 4]],
                ],
                dtype=tf.float32,
            ),
            "classes": tf.ragged.constant(
                [
                    [0, 1],
                    [2],
                ],
                dtype=tf.float32,
            ),
        }
        if is_bbox_compatible:
            try:
                layer = layer_cls(**args, bounding_box_format="xyxy")
            except TypeError:
                layer = layer_cls(**args)
            inputs = {
                IMAGES: images,
                LABELS: labels,
                BOUNDING_BOXES: bounding_boxes,
            }
        else:
            layer = layer_cls(**args)
            inputs = {IMAGES: images, LABELS: labels}

        outputs = layer(inputs)

        if layer_cls not in NO_PRESERVED_SHAPE:
            self.assertEqual(images.shape, outputs[IMAGES].shape)
        else:
            self.assertNotEqual(images.shape, outputs[IMAGES].shape)

    @parameterized.named_parameters(*GENERAL_TESTS)
    def test_layer_dtypes(self, layer_cls, args, is_bbox_compatible):
        if layer_cls in SKIP_DTYPE:
            return
        images = tf.random.uniform(shape=(2, 16, 16, 3), seed=SEED) * 255.0
        labels = tf.random.uniform(shape=(2, 1), seed=SEED) * 10.0
        bounding_boxes = {
            "boxes": tf.ragged.constant(
                [
                    [[0, 0, 2, 2], [0, 0, 16, 16]],
                    [[2, 5, 1, 4]],
                ],
                dtype=tf.float32,
            ),
            "classes": tf.ragged.constant(
                [
                    [0, 1],
                    [2],
                ],
                dtype=tf.float32,
            ),
        }
        copied_args = copy.deepcopy(args)
        if is_bbox_compatible:
            try:
                layer_cls(**copied_args, bounding_box_format="xyxy")
                copied_args["bounding_box_format"] = "xyxy"
            except TypeError:
                pass
            inputs = {
                IMAGES: images,
                LABELS: labels,
                BOUNDING_BOXES: bounding_boxes,
            }
        else:
            inputs = {IMAGES: images, LABELS: labels}

        # float64
        layer = layer_cls(**copied_args, dtype=tf.float64)
        results = layer(inputs)
        self.assertDTypeEqual(results[IMAGES], tf.float64)

        # float32
        layer = layer_cls(**copied_args)
        results = layer(inputs)
        self.assertDTypeEqual(results[IMAGES], tf.float32)

        # float16
        layer = layer_cls(**copied_args, dtype=tf.float16)
        results = layer(inputs)
        self.assertDTypeEqual(results[IMAGES], tf.float16)

        # bfloat16
        layer = layer_cls(**copied_args, dtype=tf.bfloat16)
        results = layer(inputs)
        self.assertDTypeEqual(results[IMAGES], tf.bfloat16)

        # uint8
        if layer_cls not in NO_UINT8:
            layer = layer_cls(**copied_args, dtype=tf.uint8)
            results = layer(inputs)
            self.assertDTypeEqual(results[IMAGES], tf.uint8)
        else:
            with self.assertRaises(
                (TypeError, ValueError, tf.errors.InvalidArgumentError)
            ):
                layer = layer_cls(**copied_args, dtype=tf.uint8)
                layer(inputs)

    @parameterized.named_parameters(*GENERAL_TESTS)
    def test_independence_on_batched(self, layer_cls, args, is_bbox_compatible):
        image = tf.random.uniform((16, 16, 3), seed=SEED) * 255.0
        label = tf.random.uniform((1,), seed=SEED) * 255.0
        batched_images = tf.stack((image, image, image, image, image), axis=0)
        batched_labels = tf.stack((label, label, label, label, label), axis=0)
        batched_bounding_boxes = {
            "boxes": tf.ragged.constant(
                [
                    [[10, 10, 20, 20], [100, 100, 150, 150]],
                    [[10, 10, 20, 20], [100, 100, 150, 150]],
                    [[10, 10, 20, 20], [100, 100, 150, 150]],
                    [[10, 10, 20, 20], [100, 100, 150, 150]],
                    [[10, 10, 20, 20], [100, 100, 150, 150]],
                ],
                dtype=tf.float32,
            ),
            "classes": tf.ragged.constant(
                [[0, 1], [0, 1], [0, 1], [0, 1], [0, 1]], dtype=tf.float32
            ),
        }
        copied_args = copy.deepcopy(args)
        if is_bbox_compatible:
            try:
                layer_cls(**copied_args, bounding_box_format="xyxy")
                copied_args["bounding_box_format"] = "xyxy"
            except TypeError:
                pass
            inputs = {
                IMAGES: batched_images,
                LABELS: batched_labels,
                BOUNDING_BOXES: batched_bounding_boxes,
            }
        else:
            inputs = {IMAGES: batched_images, LABELS: batched_labels}
        try:
            layer_cls(**copied_args, seed=SEED)
            copied_args["seed"] = SEED
        except TypeError:
            pass
        layer = layer_cls(**copied_args)

        results = layer(inputs)

        if layer_cls not in ALWAYS_SAME_OUTPUT_WITHIN_BATCH_LAYERS:
            self.assertNotAllClose(results[IMAGES][0], results[IMAGES][1])
        else:
            self.assertAllClose(results[IMAGES][0], results[IMAGES][1])



"""
Most of these codes come from KerasCV.
"""
import itertools

import numpy as np
import tensorflow as tf
from absl.testing import parameterized

from keras_aug.datapoints import bounding_box

xyxy_box = tf.constant(
    [[[10, 20, 110, 120], [20, 30, 120, 130]]], dtype=tf.float32
)
yxyx_box = tf.constant(
    [[[20, 10, 120, 110], [30, 20, 130, 120]]], dtype=tf.float32
)
rel_xyxy_box = tf.constant(
    [[[0.01, 0.02, 0.11, 0.12], [0.02, 0.03, 0.12, 0.13]]], dtype=tf.float32
)
rel_xyxy_box_ragged_images = tf.constant(
    [[[0.10, 0.20, 1.1, 1.20], [0.40, 0.6, 2.40, 2.6]]], dtype=tf.float32
)
rel_yxyx_box = tf.constant(
    [[[0.02, 0.01, 0.12, 0.11], [0.03, 0.02, 0.13, 0.12]]], dtype=tf.float32
)
rel_yxyx_box_ragged_images = tf.constant(
    [[[0.2, 0.1, 1.2, 1.1], [0.6, 0.4, 2.6, 2.4]]], dtype=tf.float32
)
center_xywh_box = tf.constant(
    [[[60, 70, 100, 100], [70, 80, 100, 100]]], dtype=tf.float32
)
xywh_box = tf.constant(
    [[[10, 20, 100, 100], [20, 30, 100, 100]]], dtype=tf.float32
)
rel_xywh_box = tf.constant(
    [[[0.01, 0.02, 0.1, 0.1], [0.02, 0.03, 0.1, 0.1]]], dtype=tf.float32
)
rel_xywh_box_ragged_images = tf.constant(
    [[[0.1, 0.2, 1, 1], [0.4, 0.6, 2, 2]]], dtype=tf.float32
)

ragged_images = tf.ragged.constant(
    [np.ones(shape=[100, 100, 3]), np.ones(shape=[50, 50, 3])],  # 2 images
    ragged_rank=2,
)

images = tf.ones([2, 1000, 1000, 3])

ragged_classes = tf.ragged.constant([[0], [0]], dtype=tf.float32)

boxes = {
    "xyxy": xyxy_box,
    "center_xywh": center_xywh_box,
    "rel_xywh": rel_xywh_box,
    "xywh": xywh_box,
    "rel_xyxy": rel_xyxy_box,
    "yxyx": yxyx_box,
    "rel_yxyx": rel_yxyx_box,
}

boxes_ragged_images = {
    "xyxy": xyxy_box,
    "center_xywh": center_xywh_box,
    "rel_xywh": rel_xywh_box_ragged_images,
    "xywh": xywh_box,
    "rel_xyxy": rel_xyxy_box_ragged_images,
    "yxyx": yxyx_box,
    "rel_yxyx": rel_yxyx_box_ragged_images,
}

test_cases = [
    (f"{source}_{target}", source, target)
    for (source, target) in itertools.permutations(boxes.keys(), 2)
] + [("xyxy_xyxy", "xyxy", "xyxy")]

test_image_ragged = [
    (f"{source}_{target}", source, target)
    for (source, target) in itertools.permutations(
        boxes_ragged_images.keys(), 2
    )
] + [("xyxy_xyxy", "xyxy", "xyxy")]


def _raggify(tensor):
    tensor = tf.squeeze(tensor, axis=0)
    tensor = tf.RaggedTensor.from_row_lengths(tensor, [1, 1])
    return tensor


class ConvertersTestCase(tf.test.TestCase, parameterized.TestCase):
    @parameterized.named_parameters(*test_cases)
    def test_converters(self, source, target):
        source_box = boxes[source]
        target_box = boxes[target]

        self.assertAllClose(
            bounding_box.convert_format(
                source_box, source=source, target=target, images=images
            ),
            target_box,
        )

    @parameterized.named_parameters(*test_image_ragged)
    def test_converters_ragged_images(self, source, target):
        source_box = _raggify(boxes_ragged_images[source])
        target_box = _raggify(boxes_ragged_images[target])
        self.assertAllClose(
            bounding_box.convert_format(
                source_box, source=source, target=target, images=ragged_images
            ),
            target_box,
        )

    @parameterized.named_parameters(*test_cases)
    def test_converters_unbatched(self, source, target):
        source_box = boxes[source][0]
        target_box = boxes[target][0]

        self.assertAllClose(
            bounding_box.convert_format(
                source_box, source=source, target=target, images=images[0]
            ),
            target_box,
        )

    def test_raises_with_different_image_rank(self):
        source_box = boxes["xyxy"][0]
        with self.assertRaises(ValueError):
            bounding_box.convert_format(
                source_box, source="xyxy", target="xywh", images=images
            )

    def test_without_images(self):
        source_box = boxes["xyxy"]
        target_box = boxes["xywh"]
        self.assertAllClose(
            bounding_box.convert_format(
                source_box, source="xyxy", target="xywh"
            ),
            target_box,
        )

    def test_rel_to_rel_without_images(self):
        source_box = boxes["rel_xyxy"]
        target_box = boxes["rel_yxyx"]
        self.assertAllClose(
            bounding_box.convert_format(
                source_box, source="rel_xyxy", target="rel_yxyx"
            ),
            target_box,
        )

    @parameterized.named_parameters(*test_cases)
    def test_ragged_bounding_box(self, source, target):
        source_box = _raggify(boxes[source])
        target_box = _raggify(boxes[target])
        self.assertAllClose(
            bounding_box.convert_format(
                source_box, source=source, target=target, images=images
            ),
            target_box,
        )

    @parameterized.named_parameters(*test_image_ragged)
    def test_ragged_bounding_box_ragged_images(self, source, target):
        source_box = _raggify(boxes_ragged_images[source])
        target_box = _raggify(boxes_ragged_images[target])
        self.assertAllClose(
            bounding_box.convert_format(
                source_box, source=source, target=target, images=ragged_images
            ),
            target_box,
        )

    @parameterized.named_parameters(*test_cases)
    def test_ragged_bounding_box_with_image_shape(self, source, target):
        source_box = _raggify(boxes[source])
        target_box = _raggify(boxes[target])
        self.assertAllClose(
            bounding_box.convert_format(
                source_box,
                source=source,
                target=target,
                image_shape=(1000, 1000, 3),
            ),
            target_box,
        )

    @parameterized.named_parameters(*test_image_ragged)
    def test_dense_bounding_box_with_ragged_images(self, source, target):
        source_box = _raggify(boxes_ragged_images[source])
        target_box = _raggify(boxes_ragged_images[target])
        source_bounding_boxes = {"boxes": source_box, "classes": ragged_classes}
        source_bounding_boxes = bounding_box.to_dense(source_bounding_boxes)

        result_bounding_boxes = bounding_box.convert_format(
            source_bounding_boxes,
            source=source,
            target=target,
            images=ragged_images,
        )
        result_bounding_boxes = bounding_box.to_ragged(result_bounding_boxes)

        self.assertAllClose(
            result_bounding_boxes["boxes"],
            target_box,
        )



import tensorflow as tf
from tensorflow import keras

from keras_aug.datapoints import image as image_utils
from keras_aug.layers.base.vectorized_base_random_layer import (
    VectorizedBaseRandomLayer,
)
from keras_aug.utils import augmentation as augmentation_utils


@keras.utils.register_keras_serializable(package="keras_aug")
class RandomPosterize(VectorizedBaseRandomLayer):
    """Randomly reduces the number of bits for each color channel.

    Args:
        value_range (Sequence[int|float]): The range of values the incoming
            images will have. This is typically either ``[0, 1]`` or
            ``[0, 255]`` depending on how your preprocessing pipeline is set up.
        factor (int|Sequence[int]|keras_aug.FactorSampler): The number of bits to
            keep for each channel. Must be a value between ``[1, 8]``.
            ``factor=(5, 8)`` means RandomPosterize will randomly keep 5 to 8
            bits for the image.
        seed (int|float, optional): The random seed. Defaults to
            ``None``.

    References:
        - `AutoAugment <https://arxiv.org/abs/1805.09501>`_
        - `RandAugment <https://arxiv.org/abs/1909.13719>`_
        - `KerasCV <https://github.com/keras-team/keras-cv>`_
    """  # noqa: E501

    def __init__(self, value_range, factor, seed=None, **kwargs):
        super().__init__(seed=seed, **kwargs)
        self.value_range = value_range
        if isinstance(factor, int):
            if not (0 < factor < 9):
                raise ValueError(
                    "factor value must be between [1, 8]. "
                    f"Received bits: {factor}."
                )
            factor = (factor, 8 + 1)
        elif isinstance(factor, (tuple, list)):
            factor = (factor[0], factor[1] + 1)
        self.factor = augmentation_utils.parse_factor(
            factor, min_value=0, max_value=8 + 1
        )
        self.seed = seed

    def get_random_transformation_batch(self, batch_size, **kwargs):
        # cannot sample from tf.int32 due to self.factor might be
        # NormalFactorSampler
        factors = self.factor(shape=(batch_size, 1))
        factors = tf.clip_by_value(factors, 0, 8)
        return tf.cast(factors, dtype=tf.int32)

    def augment_ragged_image(self, image, transformation, **kwargs):
        image = tf.expand_dims(image, axis=0)
        transformation = tf.expand_dims(transformation, axis=0)
        image = self.augment_images(
            images=image, transformations=transformation, **kwargs
        )
        return tf.squeeze(image, axis=0)

    def augment_images(self, images, transformations, **kwargs):
        images = image_utils.transform_value_range(
            images,
            original_range=self.value_range,
            target_range=(0, 255),
        )
        images = tf.cast(images, tf.uint8)

        inputs_for_posterize_single_image = {
            augmentation_utils.IMAGES: images,
            "bits": transformations,
        }
        images = tf.vectorized_map(
            self.posterize_single_image, inputs_for_posterize_single_image
        )
        images = tf.cast(images, self.compute_dtype)
        images = image_utils.transform_value_range(
            images,
            original_range=(0, 255),
            target_range=self.value_range,
            dtype=self.compute_dtype,
        )
        return images

    def augment_labels(self, labels, transformations, **kwargs):
        return labels

    def augment_bounding_boxes(self, bounding_boxes, transformations, **kwargs):
        return bounding_boxes

    def augment_segmentation_masks(
        self, segmentation_masks, transformations, **kwargs
    ):
        return segmentation_masks

    def augment_keypoints(self, keypoints, transformations, **kwargs):
        return keypoints

    def posterize_single_image(self, inputs):
        image = inputs.get(augmentation_utils.IMAGES, None)
        shift = 8 - tf.cast(inputs.get("bits", None), dtype=tf.uint8)
        return tf.bitwise.left_shift(
            tf.bitwise.right_shift(image, shift), shift
        )

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "value_range": self.value_range,
                "factor": self.factor,
                "seed": self.seed,
            }
        )
        return config



import tensorflow as tf
from tensorflow import keras

from keras_aug.datapoints import bounding_box
from keras_aug.datapoints import image as image_utils
from keras_aug.datapoints.bounding_box.iou import _compute_area
from keras_aug.datapoints.bounding_box.iou import _compute_intersection
from keras_aug.layers.base.vectorized_base_random_layer import (
    VectorizedBaseRandomLayer,
)
from keras_aug.utils import augmentation as augmentation_utils


@keras.utils.register_keras_serializable(package="keras_aug")
class RandomCutout(VectorizedBaseRandomLayer):
    """Randomly cut out rectangles from images and fill them.

    Args:
        height_factor (float|Sequence[float]|keras_aug.FactorSampler): The range
            of the height factor that controls the height of the cutout
            rectangle. When represented as a single float, the factor will be
            picked between ``[0.0, 0.0 + upper]``. ``0.0`` means the rectangle
            will be of size 0% of the image height. ``0.1`` means the rectangle
            will have a size of 10% of the image height.
        width_factor (float|Sequence[float]|keras_aug.FactorSampler): The range
            of the width factor that controls the width of the cutout
            rectangle. When represented as a single float, the factor will be
            picked between ``[0.0, 0.0 + upper]``. ``0.0`` means the rectangle
            will be of size 0% of the image width. ``0.1`` means the rectangle
            will have a size of 10% of the image width.
        fill_mode (str, optional): Pixels inside the cutout rectangle are filled
            according to the given mode. Supported values:
            ``"constant", "gaussian_noise"``. Defaults to ``"constant"``.
        fill_value (int|float, optional): The value to be filled in the cutout
            rectangle when ``fill_mode="constant"``. Defaults to ``0``.
        bbox_removal_threshold (float, optional): The bounding boxes having
            content cut above the threshold will be removed.
            Defaults to ``0.6`` which is applied by ultralytics/yolo series.
        bounding_box_format (str, optional): The format of bounding
            boxes of input dataset. Refer
            https://github.com/james77777778/keras-aug/blob/main/keras_aug/datapoints/bounding_box/converter.py
            for more details on supported bounding box formats.
        seed (int|float, optional): The random seed. Defaults to ``None``.

    References:
        - `Cutout <https://arxiv.org/abs/1708.04552>`_
        - `KerasCV <https://github.com/keras-team/keras-cv>`_
        - `kaushal2896@Kaggle <https://www.kaggle.com/code/kaushal2896/data-augmentation-tutorial-basic-cutout-mixup>`_
        - `ultralytics/ultralytics <https://github.com/ultralytics/ultralytics>`_
    """  # noqa: E501

    def __init__(
        self,
        height_factor,
        width_factor,
        fill_mode="constant",
        fill_value=0.0,
        bbox_removal_threshold=0.6,
        bounding_box_format=None,
        seed=None,
        **kwargs,
    ):
        super().__init__(seed=seed, **kwargs)
        if isinstance(height_factor, (int, float)):
            height_factor = (0, height_factor)
        self.height_factor = augmentation_utils.parse_factor(
            height_factor, seed=seed
        )
        if isinstance(width_factor, (int, float)):
            width_factor = (0, width_factor)
        self.width_factor = augmentation_utils.parse_factor(
            width_factor, seed=seed
        )
        if fill_mode not in ["gaussian_noise", "constant"]:
            raise ValueError(
                '`fill_mode` should be "gaussian_noise" '
                f'or "constant". Got `fill_mode`={fill_mode}'
            )
        self.fill_mode = fill_mode
        self.fill_value = fill_value
        self.bbox_removal_threshold = bbox_removal_threshold
        self.bounding_box_format = bounding_box_format
        self.seed = seed

    def get_random_transformation_batch(
        self, batch_size, images=None, **kwargs
    ):
        heights, widths = augmentation_utils.get_images_shape(
            images, dtype=tf.float32
        )
        center_xs = self._random_generator.random_uniform(
            shape=(batch_size, 1), minval=0, maxval=1, dtype=tf.float32
        )
        center_ys = self._random_generator.random_uniform(
            shape=(batch_size, 1), minval=0, maxval=1, dtype=tf.float32
        )
        center_xs = tf.cast(tf.round(center_xs * widths), dtype=tf.int32)
        center_ys = tf.cast(tf.round(center_ys * heights), dtype=tf.int32)

        cutout_heights = self.height_factor(shape=(batch_size, 1))
        cutout_widths = self.width_factor(shape=(batch_size, 1))
        cutout_heights = cutout_heights * heights
        cutout_widths = cutout_widths * widths
        cutout_heights = tf.cast(tf.math.ceil(cutout_heights), tf.int32)
        cutout_widths = tf.cast(tf.math.ceil(cutout_widths), tf.int32)
        cutout_heights = tf.minimum(
            cutout_heights, tf.cast(heights, dtype=tf.int32)
        )
        cutout_widths = tf.minimum(
            cutout_widths, tf.cast(widths, dtype=tf.int32)
        )
        return {
            "center_xs": center_xs,
            "center_ys": center_ys,
            "cutout_heights": cutout_heights,
            "cutout_widths": cutout_widths,
        }

    def augment_ragged_image(self, image, transformation, **kwargs):
        image = tf.expand_dims(image, axis=0)
        transformation = augmentation_utils.expand_dict_dims(
            transformation, axis=0
        )
        image = self.augment_images(
            images=image, transformations=transformation, **kwargs
        )
        return tf.squeeze(image, axis=0)

    def augment_images(self, images, transformations, **kwargs):
        images = tf.cast(images, dtype=self.compute_dtype)
        center_xs = transformations["center_xs"]
        center_ys = transformations["center_ys"]
        cutout_heights = transformations["cutout_heights"]
        cutout_widths = transformations["cutout_widths"]
        rectangle_fills = self.compute_rectangle_fills(images)
        images = image_utils.fill_rectangle(
            images,
            center_xs[..., 0],
            center_ys[..., 0],
            cutout_widths[..., 0],
            cutout_heights[..., 0],
            rectangle_fills,
        )
        return images

    def augment_labels(self, labels, transformations, **kwargs):
        return labels

    def augment_bounding_boxes(
        self, bounding_boxes, transformations, raw_images=None, **kwargs
    ):
        if self.bounding_box_format is None:
            raise ValueError(
                "`RandomCutout()` was called with bounding boxes,"
                "but no `bounding_box_format` was specified in the constructor."
                "Please specify a bounding box format in the constructor. i.e."
                "`RandomCutout(bounding_box_format='xyxy')`"
            )
        bounding_boxes = bounding_box.to_dense(bounding_boxes)
        bounding_boxes = bounding_box.convert_format(
            bounding_boxes,
            source=self.bounding_box_format,
            target="xyxy",
            images=raw_images,
            dtype=tf.float32,
        )
        # construct cutout bounding_boxes
        cutout_boxes = tf.concat(
            [
                transformations["center_xs"],
                transformations["center_ys"],
                transformations["cutout_widths"],
                transformations["cutout_heights"],
            ],
            axis=-1,
        )
        cutout_boxes = tf.expand_dims(cutout_boxes, axis=1)
        cutout_boxes = bounding_box.convert_format(
            cutout_boxes,
            source="center_xywh",
            target="xyxy",
            images=raw_images,
            dtype=tf.float32,
        )

        areas = _compute_area(bounding_boxes["boxes"])
        intersections = tf.squeeze(
            _compute_intersection(bounding_boxes["boxes"], cutout_boxes),
            axis=-1,
        )
        intersection_ratios = tf.math.divide_no_nan(intersections, areas)

        # set classes == -1 if ratios < self.bbox_removal_threshold
        # the bounding_boxes with classes==-1 will be removed by
        # bounding_box.to_ragged() after self.augment_bounding_boxes()
        bounding_boxes["classes"] = tf.where(
            intersection_ratios >= self.bbox_removal_threshold,
            -1.0,
            bounding_boxes["classes"],
        )
        return bounding_boxes

    def compute_rectangle_fills(self, inputs):
        input_shape = tf.shape(inputs)
        if self.fill_mode == "constant":
            fill_value = tf.fill(input_shape, self.fill_value)
            fill_value = tf.cast(fill_value, dtype=self.compute_dtype)
        else:
            # gaussian noise
            fill_value = tf.random.normal(input_shape, dtype=self.compute_dtype)
        return fill_value

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "height_factor": self.height_factor,
                "width_factor": self.width_factor,
                "fill_mode": self.fill_mode,
                "fill_value": self.fill_value,
                "bbox_removal_threshold": self.bbox_removal_threshold,
                "bounding_box_format": self.bounding_box_format,
                "seed": self.seed,
            }
        )
        return config



import enum
from typing import Sequence

import tensorflow as tf

from keras_aug.core import ConstantFactorSampler
from keras_aug.core import FactorSampler
from keras_aug.core import NormalFactorSampler
from keras_aug.core import SignedNormalFactorSampler
from keras_aug.core import UniformFactorSampler

H_AXIS = -3
W_AXIS = -2

IMAGES = "images"
LABELS = "labels"
TARGETS = "targets"
BOUNDING_BOXES = "bounding_boxes"
KEYPOINTS = "keypoints"
SEGMENTATION_MASKS = "segmentation_masks"
CUSTOM_ANNOTATIONS = "custom_annotations"

BATCHED = "batched"


class PaddingPosition(enum.Enum):
    CENTER = "center"
    TOP_LEFT = "top_left"
    TOP_RIGHT = "top_right"
    BOTTOM_LEFT = "bottom_left"
    BOTTOM_RIGHT = "bottom_right"
    RANDOM = "random"


def get_padding_position(position):
    if isinstance(position, PaddingPosition):
        return position
    position = position.lower()
    if position not in PaddingPosition._value2member_map_.keys():
        raise NotImplementedError(
            f"Value not recognized for `position`: {position}. Supported "
            f"values are: {PaddingPosition._value2member_map_.keys()}"
        )
    return PaddingPosition._value2member_map_[position]


def get_position_params(
    tops, bottoms, lefts, rights, position, random_generator
):
    """This function supposes arguments are at `center` padding method."""
    tops = tf.convert_to_tensor(tops)
    bottoms = tf.convert_to_tensor(bottoms)
    lefts = tf.convert_to_tensor(lefts)
    rights = tf.convert_to_tensor(rights)

    if position == PaddingPosition.CENTER:
        # do nothing
        bottoms = bottoms
        rights = rights
        tops = tops
        lefts = lefts
    elif position == PaddingPosition.TOP_LEFT:
        bottoms += tops
        rights += lefts
        tops = tf.zeros_like(tops)
        lefts = tf.zeros_like(lefts)
    elif position == PaddingPosition.TOP_RIGHT:
        bottoms += tops
        lefts += rights
        tops = tf.zeros_like(tops)
        rights = tf.zeros_like(rights)
    elif position == PaddingPosition.BOTTOM_LEFT:
        tops += bottoms
        rights += lefts
        bottoms = tf.zeros_like(bottoms)
        lefts = tf.zeros_like(lefts)
    elif position == PaddingPosition.BOTTOM_RIGHT:
        tops += bottoms
        lefts += rights
        bottoms = tf.zeros_like(bottoms)
        rights = tf.zeros_like(rights)
    elif position == PaddingPosition.RANDOM:
        batch_size = tf.shape(tops)[0]
        original_dtype = tops.dtype
        h_pads = tf.cast(tops + bottoms, dtype=tf.float32)
        w_pads = tf.cast(lefts + rights, dtype=tf.float32)
        tops = random_generator.random_uniform(
            shape=(batch_size, 1), minval=0, maxval=1, dtype=tf.float32
        )
        tops = tf.cast(tf.round(tops * h_pads), dtype=original_dtype)
        bottoms = tf.cast(h_pads, dtype=tf.int32) - tops
        lefts = random_generator.random_uniform(
            shape=(batch_size, 1), minval=0, maxval=1, dtype=tf.float32
        )
        lefts = tf.cast(tf.round(lefts * w_pads), dtype=original_dtype)
        rights = tf.cast(w_pads, dtype=tf.int32) - lefts
    else:
        raise NotImplementedError(
            f"Value not recognized for `position`: {position}. Supported "
            f"values are: {PaddingPosition._value2member_map_.keys()}"
        )

    return tops, bottoms, lefts, rights


def parse_factor(
    param,
    min_value=0.0,
    max_value=1.0,
    center_value=0.5,
    param_name="factor",
    seed=None,
):
    if isinstance(param, FactorSampler):
        return param

    if isinstance(param, float) or isinstance(param, int):
        param = (center_value - param, center_value + param)

    if param[0] > param[1]:
        raise ValueError(
            f"`{param_name}[0] > {param_name}[1]`, `{param_name}[0]` must be "
            f"<= `{param_name}[1]`. Got `{param_name}={param}`"
        )
    if (min_value is not None and param[0] < min_value) or (
        max_value is not None and param[1] > max_value
    ):
        raise ValueError(
            f"`{param_name}` should be inside of range "
            f"[{min_value}, {max_value}]. Got {param_name}={param}"
        )

    if param[0] == param[1]:
        return ConstantFactorSampler(param[0])

    return UniformFactorSampler(param[0], param[1], seed=seed)


def is_factor_working(factor, not_working_value=0.0):
    """Check whether ``factor`` is working or not.

    Args:
        factor (int|float|Sequence[int|float]|keras_aug.FactorSampler): The
            factor to check whether it is working or not.
        not_working_value (float, optional): The value indicating not working
            status. Defaults to ``0.0``.
    """
    if factor is None:
        return False
    if isinstance(factor, (int, float)):
        if factor == not_working_value:
            return False
    elif isinstance(factor, Sequence):
        if factor[0] == factor[1] and factor[0] == not_working_value:
            return False
    elif isinstance(factor, ConstantFactorSampler):
        if factor.value == not_working_value:
            return False
    elif isinstance(factor, UniformFactorSampler):
        if (
            factor.lower == not_working_value
            and factor.upper == not_working_value
        ):
            return False
    elif isinstance(factor, (NormalFactorSampler, SignedNormalFactorSampler)):
        if factor.stddev == 0 and factor.mean == not_working_value:
            return False
    else:
        raise ValueError(
            f"Cannot recognize factor type: {factor} with type {type(factor)}"
        )
    return True


def expand_dict_dims(dicts, axis):
    new_dicts = {}
    for key in dicts.keys():
        tensor = dicts[key]
        new_dicts[key] = tf.expand_dims(tensor, axis=axis)
    return new_dicts


def get_images_shape(images, dtype=tf.int32):
    """Get ``heights`` and ``widths`` of the input images.

    Input images can be ``tf.Tensor`` or ``tf.RaggedTensor`` with the shape of
    [B, H|None, W|None, C].

    Args:
        images (tf.Tensor|tf.RaggedTensor): The input images.
        dtype (tf.dtypes.DType, optional): The dtype of the outputs. Defaults to
            ``tf.int32``.
    """
    if isinstance(images, tf.RaggedTensor):
        heights = tf.reshape(images.row_lengths(), (-1, 1))
        widths = tf.reshape(
            tf.reduce_max(images.row_lengths(axis=2), 1), (-1, 1)
        )
    else:
        batch_size = tf.shape(images)[0]
        heights = tf.repeat(tf.shape(images)[H_AXIS], repeats=[batch_size])
        heights = tf.reshape(heights, shape=(-1, 1))
        widths = tf.repeat(tf.shape(images)[W_AXIS], repeats=[batch_size])
        widths = tf.reshape(widths, shape=(-1, 1))
    return tf.cast(heights, dtype=dtype), tf.cast(widths, dtype=dtype)


def cast_to(inputs, dtype):
    if IMAGES in inputs:
        inputs[IMAGES] = tf.cast(inputs[IMAGES], dtype)
    if LABELS in inputs:
        inputs[LABELS] = tf.cast(inputs[LABELS], dtype)
    if BOUNDING_BOXES in inputs:
        inputs[BOUNDING_BOXES]["boxes"] = tf.cast(
            inputs[BOUNDING_BOXES]["boxes"], dtype
        )
        inputs[BOUNDING_BOXES]["classes"] = tf.cast(
            inputs[BOUNDING_BOXES]["classes"], dtype
        )
    if SEGMENTATION_MASKS in inputs:
        inputs[SEGMENTATION_MASKS] = tf.cast(inputs[SEGMENTATION_MASKS], dtype)
    if KEYPOINTS in inputs:
        inputs[KEYPOINTS] = tf.cast(inputs[KEYPOINTS], dtype)
    if CUSTOM_ANNOTATIONS in inputs:
        raise NotImplementedError()
    return inputs


def compute_signature(inputs, dtype):
    fn_output_signature = {}
    if IMAGES in inputs:
        if isinstance(inputs[IMAGES], tf.Tensor):
            fn_output_signature[IMAGES] = tf.TensorSpec(
                inputs[IMAGES].shape[1:], dtype
            )
        else:
            fn_output_signature[IMAGES] = tf.RaggedTensorSpec(
                shape=inputs[IMAGES].shape[1:],
                ragged_rank=1,
                dtype=dtype,
            )
    if LABELS in inputs:
        fn_output_signature[LABELS] = tf.TensorSpec(
            inputs[LABELS].shape[1:], dtype
        )
    if BOUNDING_BOXES in inputs:
        fn_output_signature[BOUNDING_BOXES] = {
            "boxes": tf.RaggedTensorSpec(
                shape=[None, 4],
                ragged_rank=1,
                dtype=dtype,
            ),
            "classes": tf.RaggedTensorSpec(
                shape=[None], ragged_rank=0, dtype=dtype
            ),
        }
    if SEGMENTATION_MASKS in inputs:
        if isinstance(inputs[SEGMENTATION_MASKS], tf.Tensor):
            fn_output_signature[SEGMENTATION_MASKS] = tf.TensorSpec(
                inputs[SEGMENTATION_MASKS].shape[1:], dtype
            )
        else:
            fn_output_signature[SEGMENTATION_MASKS] = tf.RaggedTensorSpec(
                shape=inputs[SEGMENTATION_MASKS].shape[1:],
                ragged_rank=1,
                dtype=dtype,
            )
    if KEYPOINTS in inputs:
        if isinstance(inputs[KEYPOINTS], tf.Tensor):
            fn_output_signature[KEYPOINTS] = tf.TensorSpec(
                inputs[KEYPOINTS].shape[1:], dtype
            )
        else:
            fn_output_signature[KEYPOINTS] = tf.RaggedTensorSpec(
                shape=inputs[KEYPOINTS].shape[1:],
                ragged_rank=1,
                dtype=dtype,
            )
    if CUSTOM_ANNOTATIONS in inputs:
        raise NotImplementedError()
    return fn_output_signature


def random_inversion(rng):
    negate = rng.random_uniform((), 0, 1, dtype=tf.float32) > 0.5
    negate = tf.cond(negate, lambda: -1.0, lambda: 1.0)
    return negate


def ensure_tensor(inputs, dtype=None):
    """Ensures the input is a Tensor, SparseTensor or RaggedTensor."""
    if not isinstance(inputs, (tf.Tensor, tf.RaggedTensor, tf.SparseTensor)):
        inputs = tf.convert_to_tensor(inputs, dtype)
    if dtype is not None and inputs.dtype != dtype:
        inputs = tf.cast(inputs, dtype)
    return inputs


_TF_INTERPOLATION_METHODS = {
    "bilinear": tf.image.ResizeMethod.BILINEAR,
    "nearest": tf.image.ResizeMethod.NEAREST_NEIGHBOR,
    "bicubic": tf.image.ResizeMethod.BICUBIC,
    "area": tf.image.ResizeMethod.AREA,
    "lanczos3": tf.image.ResizeMethod.LANCZOS3,
    "lanczos5": tf.image.ResizeMethod.LANCZOS5,
    "gaussian": tf.image.ResizeMethod.GAUSSIAN,
    "mitchellcubic": tf.image.ResizeMethod.MITCHELLCUBIC,
}


def get_interpolation(interpolation):
    interpolation = interpolation.lower()
    if interpolation not in _TF_INTERPOLATION_METHODS:
        raise NotImplementedError(
            "Value not recognized for `interpolation`: {}. Supported values "
            "are: {}".format(interpolation, _TF_INTERPOLATION_METHODS.keys())
        )
    return _TF_INTERPOLATION_METHODS[interpolation]


def check_fill_mode_and_interpolation(fill_mode, interpolation):
    if fill_mode not in {"reflect", "wrap", "constant", "nearest"}:
        raise NotImplementedError(
            " Want fillmode  to be one of `reflect`, `wrap`, "
            "`constant` or `nearest`. Got `fill_mode` {}. ".format(fill_mode)
        )
    if interpolation not in {"nearest", "bilinear"}:
        raise NotImplementedError(
            "Unknown `interpolation` {}. Only `nearest` and "
            "`bilinear` are supported.".format(interpolation)
        )



import tempfile

import tensorflow as tf

from keras_aug.core.keras_random_generator import KerasRandomGenerator
from keras_aug.layers.__internal__.base_layer import BaseRandomLayer


class BaseRandomLayerTest(tf.test.TestCase):
    def test_saved_model(self):
        base_random_layer = BaseRandomLayer(
            force_generator=True, rng_type="stateless"
        )

        with tempfile.TemporaryDirectory() as tmpdirname:
            tf.saved_model.save(base_random_layer, tmpdirname)
            tf.saved_model.load(tmpdirname)

    def test_checkpoint(self):
        inputs = tf.keras.Input(shape=(1,))
        base_random_layer = BaseRandomLayer(
            force_generator=True, rng_type="stateless"
        )
        outputs = base_random_layer(inputs)
        model = tf.keras.Model(inputs, outputs)
        checkpoint = tf.train.Checkpoint(model)

        with tempfile.TemporaryDirectory() as tmpdirname:
            save_path = checkpoint.save(tmpdirname)
            checkpoint.restore(save_path)

    def test_lookup_dependency(self):
        base_random_layer = BaseRandomLayer(
            force_generator=True, rng_type="stateless"
        )

        rng = base_random_layer._lookup_dependency("_random_generator")

        self.assertTrue(isinstance(rng, KerasRandomGenerator))



import tensorflow as tf
from tensorflow import keras

from keras_aug.core.factor_sampler.factor_sampler import FactorSampler


@keras.utils.register_keras_serializable(package="keras_aug")
class ConstantFactorSampler(FactorSampler):
    """ConstantFactorSampler samples the same factor for every call to
    ``__call__()``.

    Args:
        value (int|float): the value to return from ``__call__()``.

    References:
        - `KerasCV <https://github.com/keras-team/keras-cv>`_
    """

    def __init__(self, value):
        self.value = value

    def __call__(self, shape=(), dtype="float32"):
        return tf.ones(shape=shape, dtype=dtype) * self.value

    def get_config(self):
        return {"value": self.value}



from keras_aug.layers.augmentation.auto.aug_mix import AugMix
from keras_aug.layers.augmentation.auto.rand_augment import RandAugment
from keras_aug.layers.augmentation.auto.trivial_augment_wide import (
    TrivialAugmentWide,
)
from keras_aug.layers.augmentation.geometry.random_affine import RandomAffine
from keras_aug.layers.augmentation.geometry.random_crop import RandomCrop
from keras_aug.layers.augmentation.geometry.random_crop_and_resize import (
    RandomCropAndResize,
)
from keras_aug.layers.augmentation.geometry.random_flip import RandomFlip
from keras_aug.layers.augmentation.geometry.random_resize import RandomResize
from keras_aug.layers.augmentation.geometry.random_rotate import RandomRotate
from keras_aug.layers.augmentation.geometry.random_zoom_and_crop import (
    RandomZoomAndCrop,
)
from keras_aug.layers.augmentation.intensity.channel_shuffle import (
    ChannelShuffle,
)
from keras_aug.layers.augmentation.intensity.random_blur import RandomBlur
from keras_aug.layers.augmentation.intensity.random_channel_shift import (
    RandomChannelShift,
)
from keras_aug.layers.augmentation.intensity.random_clahe import RandomCLAHE
from keras_aug.layers.augmentation.intensity.random_color_jitter import (
    RandomColorJitter,
)
from keras_aug.layers.augmentation.intensity.random_gamma import RandomGamma
from keras_aug.layers.augmentation.intensity.random_gaussian_blur import (
    RandomGaussianBlur,
)
from keras_aug.layers.augmentation.intensity.random_hsv import RandomHSV
from keras_aug.layers.augmentation.intensity.random_jpeg_quality import (
    RandomJpegQuality,
)
from keras_aug.layers.augmentation.intensity.random_posterize import (
    RandomPosterize,
)
from keras_aug.layers.augmentation.intensity.random_sharpness import (
    RandomSharpness,
)
from keras_aug.layers.augmentation.intensity.random_solarize import (
    RandomSolarize,
)
from keras_aug.layers.augmentation.mix.cut_mix import CutMix
from keras_aug.layers.augmentation.mix.mix_up import MixUp
from keras_aug.layers.augmentation.mix.mosaic import Mosaic
from keras_aug.layers.augmentation.regularization.random_channel_dropout import (  # noqa: E501
    RandomChannelDropout,
)
from keras_aug.layers.augmentation.regularization.random_cutout import (
    RandomCutout,
)
from keras_aug.layers.augmentation.regularization.random_erase import (
    RandomErase,
)
from keras_aug.layers.augmentation.regularization.random_grid_mask import (
    RandomGridMask,
)
from keras_aug.layers.augmentation.utility.random_apply import RandomApply
from keras_aug.layers.augmentation.utility.random_choice import RandomChoice
from keras_aug.layers.augmentation.utility.repeated_augment import (
    RepeatedAugment,
)



import tensorflow as tf
from tensorflow import keras

from keras_aug.layers.base.vectorized_base_random_layer import (
    VectorizedBaseRandomLayer,
)


@keras.utils.register_keras_serializable(package="keras_aug")
class Invert(VectorizedBaseRandomLayer):
    """Inverts the inputs.

    Inverts the pixel value by equation: ``y = max_pixel_value - x``.

    Args:
        value_range (Sequence[int|float]): The range of values the incoming
            images will have. This is typically either ``[0, 1]`` or
            ``[0, 255]`` depending on how your preprocessing pipeline is set up.
    """

    def __init__(self, value_range, **kwargs):
        super().__init__(**kwargs)
        self.value_range = value_range

    def augment_ragged_image(self, image, transformation, **kwargs):
        return self.augment_images(
            images=image, transformations=transformation, **kwargs
        )

    def augment_images(self, images, transformations, **kwargs):
        images = tf.cast(images, dtype=self.compute_dtype)
        return self.value_range[1] - images

    def augment_labels(self, labels, transformations, **kwargs):
        return labels

    def augment_bounding_boxes(self, bounding_boxes, transformations, **kwargs):
        return bounding_boxes

    def augment_segmentation_masks(
        self, segmentation_masks, transformations, **kwargs
    ):
        return segmentation_masks

    def augment_keypoints(self, keypoints, transformations, **kwargs):
        return keypoints

    def get_config(self):
        config = super().get_config()
        config.update({"value_range": self.value_range})
        return config



from keras_aug.layers.preprocessing.geometry.center_crop import CenterCrop
from keras_aug.layers.preprocessing.geometry.pad_if_needed import PadIfNeeded
from keras_aug.layers.preprocessing.geometry.resize import Resize
from keras_aug.layers.preprocessing.intensity.auto_contrast import AutoContrast
from keras_aug.layers.preprocessing.intensity.equalize import Equalize
from keras_aug.layers.preprocessing.intensity.grayscale import Grayscale
from keras_aug.layers.preprocessing.intensity.identity import Identity
from keras_aug.layers.preprocessing.intensity.invert import Invert
from keras_aug.layers.preprocessing.intensity.normalize import Normalize
from keras_aug.layers.preprocessing.intensity.rescale import Rescale
from keras_aug.layers.preprocessing.utility.sanitize_bounding_box import (
    SanitizeBoundingBox,
)



import tensorflow as tf
from tensorflow import keras

from keras_aug.datapoints import image as image_utils
from keras_aug.layers.base.vectorized_base_random_layer import (
    VectorizedBaseRandomLayer,
)
from keras_aug.utils import augmentation as augmentation_utils


@keras.utils.register_keras_serializable(package="keras_aug")
class RandomSolarize(VectorizedBaseRandomLayer):
    """Randomly applies ``(max_value - pixel + min_value)`` for each pixel in
    the input images.

    When created without ``threshold_factor`` parameter, the layer performs
    solarization to all values. When created with specified ``threshold_factor``
    the layer only augments pixels that are above the ``threshold_factor``
    value.

    Args:
        value_range (Sequence[int|float]): The range of values the incoming
            images will have. This is typically either ``[0, 1]`` or
            ``[0, 255]`` depending on how your preprocessing pipeline is set up.
        threshold_factor (float|Sequence[float]|keras_aug.FactorSampler):
            The range of the threshold factor. Only the pixel values above the
            threshold will be solarized. When represented as a single float,
            the factor will be picked between ``[0, upper]``. ``255``
            means no thresholding.
        addition_factor (float|Sequence[float]|keras_aug.FactorSampler, optional):
            The range of the addition factor that is added to each pixel before
            solarization and thresholding. When represented as a single float,
            the factor will be picked between ``[0, upper]``. ``0`` means no
            addition. Defaults to ``0``.
        seed (int|float, optional): The random seed. Defaults to
            ``None``.

    References:
        - `AutoAugment <https://arxiv.org/abs/1805.09501>`_
        - `RandAugment <https://arxiv.org/abs/1909.13719>`_
        - `KerasCV <https://github.com/keras-team/keras-cv>`_
    """  # noqa: E501

    def __init__(
        self,
        value_range,
        threshold_factor,
        addition_factor=0,
        seed=None,
        **kwargs,
    ):
        super().__init__(seed=seed, **kwargs)
        self.value_range = value_range
        if isinstance(threshold_factor, (int, float)):
            threshold_factor = (0, threshold_factor)
        self.threshold_factor = augmentation_utils.parse_factor(
            threshold_factor,
            max_value=255,
            seed=seed,
            param_name="threshold_factor",
        )
        if isinstance(addition_factor, (int, float)):
            addition_factor = (0, addition_factor)
        self.addition_factor = augmentation_utils.parse_factor(
            addition_factor,
            max_value=255,
            seed=seed,
            param_name="addition_factor",
        )
        self.seed = seed

    def get_random_transformation_batch(self, batch_size, **kwargs):
        return {
            "additions": self.addition_factor(
                shape=(batch_size, 1, 1, 1), dtype=self.compute_dtype
            ),
            "thresholds": self.threshold_factor(
                shape=(batch_size, 1, 1, 1), dtype=self.compute_dtype
            ),
        }

    def augment_ragged_image(self, image, transformation, **kwargs):
        image = tf.expand_dims(image, axis=0)
        transformation = augmentation_utils.expand_dict_dims(
            transformation, axis=0
        )
        image = self.augment_images(
            images=image, transformations=transformation, **kwargs
        )
        return tf.squeeze(image, axis=0)

    def augment_images(self, images, transformations, **kwargs):
        thresholds = transformations["thresholds"]
        additions = transformations["additions"]
        images = image_utils.transform_value_range(
            images,
            original_range=self.value_range,
            target_range=(0, 255),
            dtype=self.compute_dtype,
        )
        images = images + additions
        images = tf.clip_by_value(images, 0, 255)
        images = tf.where(images < thresholds, images, 255 - images)
        images = image_utils.transform_value_range(
            images,
            original_range=(0, 255),
            target_range=self.value_range,
            dtype=self.compute_dtype,
        )
        return images

    def augment_labels(self, labels, transformations, **kwargs):
        return labels

    def augment_bounding_boxes(self, bounding_boxes, transformations, **kwargs):
        return bounding_boxes

    def augment_segmentation_masks(
        self, segmentation_masks, transformations, **kwargs
    ):
        return segmentation_masks

    def augment_keypoints(self, keypoints, transformations, **kwargs):
        return keypoints

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "value_range": self.value_range,
                "threshold_factor": self.threshold_factor,
                "addition_factor": self.addition_factor,
                "seed": self.seed,
            }
        )
        return config



import tensorflow as tf
from absl.testing import parameterized

from keras_aug import layers
from keras_aug.datapoints import bounding_box


class RandomZoomAndCropTest(tf.test.TestCase, parameterized.TestCase):
    batch_size = 4
    ori_height = 9
    ori_width = 8
    seed = 13
    height = 4
    width = 4

    def test_train_augments_image(self):
        # Checks if original and augmented images are different

        input_image_shape = (
            self.batch_size,
            self.ori_height,
            self.ori_width,
            3,
        )
        image = tf.random.uniform(shape=input_image_shape, seed=self.seed)

        layer = layers.RandomZoomAndCrop(
            height=self.height,
            width=self.width,
            scale_factor=(3 / 4, 4 / 3),
            seed=self.seed,
        )
        output = layer(image)

        input_image_resized = tf.image.resize(image, (self.height, self.width))

        self.assertNotAllClose(output, input_image_resized)

    def test_augment_bounding_box_single(self):
        image = tf.zeros([20, 20, 3])
        boxes = {
            "boxes": tf.convert_to_tensor([[0, 0, 1, 1]], dtype=tf.float32),
            "classes": tf.convert_to_tensor([0], dtype=tf.float32),
        }
        input = {"images": image, "bounding_boxes": boxes}

        layer = layers.RandomZoomAndCrop(
            height=self.height,
            width=self.width,
            scale_factor=(3 / 4, 4 / 3),
            bounding_box_format="rel_xyxy",
            seed=self.seed,
        )
        output = layer(input)
        output["bounding_boxes"] = bounding_box.to_dense(
            output["bounding_boxes"]
        )
        expected_output = {
            "boxes": tf.convert_to_tensor([[0, 0, 1, 1]], dtype=tf.float32),
            "classes": tf.convert_to_tensor([0], dtype=tf.float32),
        }
        self.assertAllClose(
            expected_output["boxes"],
            output["bounding_boxes"]["boxes"],
        )
        self.assertAllClose(
            expected_output["classes"], output["bounding_boxes"]["classes"]
        )

    def test_augment_boxes_batched_input(self):
        image = tf.zeros([20, 20, 3])

        bounding_boxes = {
            "classes": tf.convert_to_tensor([[0, 0], [0, 0]]),
            "boxes": tf.convert_to_tensor(
                [
                    [[0, 0, 1, 1], [0, 0, 1, 1]],
                    [[0, 0, 1, 1], [0, 0, 1, 1]],
                ]
            ),
        }
        input = {"images": [image, image], "bounding_boxes": bounding_boxes}

        layer = layers.RandomZoomAndCrop(
            height=self.height,
            width=self.width,
            scale_factor=(3 / 4, 4 / 3),
            bounding_box_format="rel_xyxy",
            seed=self.seed,
        )
        output = layer(input, training=True)
        output["bounding_boxes"] = bounding_box.to_dense(
            output["bounding_boxes"]
        )
        expected_output = {
            "classes": tf.convert_to_tensor([[0, 0], [0, 0]], dtype=tf.float32),
            "boxes": tf.convert_to_tensor(
                [
                    [[0, 0, 1, 1], [0, 0, 1, 1]],
                    [[0, 0, 1, 1], [0, 0, 1, 1]],
                ],
                dtype=tf.float32,
            ),
        }
        self.assertAllClose(
            expected_output["boxes"],
            output["bounding_boxes"]["boxes"],
        )
        self.assertAllClose(
            expected_output["classes"], output["bounding_boxes"]["classes"]
        )

    def test_augment_boxes_ragged(self):
        image = tf.zeros([2, 20, 20, 3])
        boxes = {
            "boxes": tf.ragged.constant(
                [[[0, 0, 1, 1], [0, 0, 1, 1]], [[0, 0, 1, 1]]], dtype=tf.float32
            ),
            "classes": tf.ragged.constant(
                [[0, 0], [0]],
                dtype=tf.float32,
            ),
        }
        input = {"images": image, "bounding_boxes": boxes}

        layer = layers.RandomZoomAndCrop(
            height=self.height,
            width=self.width,
            scale_factor=(3 / 4, 4 / 3),
            bounding_box_format="rel_xyxy",
            seed=self.seed,
        )
        output = layer(input)
        # the result boxes will still have the entire image in them
        expected_output = {
            "boxes": tf.ragged.constant(
                [[[0, 0, 1, 1], [0, 0, 1, 1]], [[0, 0, 1, 1]]], dtype=tf.float32
            ),
            "classes": tf.ragged.constant(
                [[0.0, 0.0], [0.0]],
                dtype=tf.float32,
            ),
        }
        self.assertAllClose(
            expected_output["boxes"].to_tensor(),
            output["bounding_boxes"]["boxes"].to_tensor(),
        )
        self.assertAllClose(
            expected_output["classes"], output["bounding_boxes"]["classes"]
        )

    def test_dense_segmentation_masks(self):
        images = tf.random.uniform((2, 10, 10, 3))
        segmentation_masks = tf.random.uniform(
            (2, 10, 10, 1), minval=0, maxval=10, dtype=tf.int32
        )
        layer = layers.RandomZoomAndCrop(
            height=self.height,
            width=self.width,
            scale_factor=(3 / 4, 4 / 3),
            bounding_box_format="rel_xyxy",
            seed=self.seed,
        )

        result = layer(
            {"images": images, "segmentation_masks": segmentation_masks}
        )

        self.assertTrue(isinstance(result["segmentation_masks"], tf.Tensor))
        self.assertEqual(
            result["segmentation_masks"].shape[1:3], (self.height, self.width)
        )
        self.assertAllInSet(result["segmentation_masks"], tf.range(0, 10))

    def test_ragged_segmentation_masks(self):
        images = tf.ragged.stack(
            [
                tf.random.uniform((8, 8, 3), dtype=tf.float32),
                tf.random.uniform((16, 8, 3), dtype=tf.float32),
            ]
        )
        segmentation_masks = tf.ragged.stack(
            [
                tf.random.uniform((8, 8, 1), maxval=10, dtype=tf.int32),
                tf.random.uniform((16, 8, 1), maxval=10, dtype=tf.int32),
            ]
        )
        segmentation_masks = tf.cast(segmentation_masks, dtype=tf.float32)
        layer = layers.RandomZoomAndCrop(
            height=self.height,
            width=self.width,
            scale_factor=(3 / 4, 4 / 3),
            bounding_box_format="rel_xyxy",
            seed=self.seed,
        )

        result = layer(
            {"images": images, "segmentation_masks": segmentation_masks}
        )

        self.assertTrue(isinstance(result["segmentation_masks"], tf.Tensor))
        self.assertEqual(
            result["segmentation_masks"].shape[1:3], (self.height, self.width)
        )
        self.assertAllInSet(result["segmentation_masks"], tf.range(0, 10))

    def test_ragged_input_with_graph_mode(self):
        images = tf.ragged.stack(
            [
                tf.random.uniform((8, 8, 3), dtype=tf.float32),
                tf.random.uniform((16, 8, 3), dtype=tf.float32),
            ]
        )
        segmentation_masks = tf.ragged.stack(
            [
                tf.random.uniform((8, 8, 1), maxval=10, dtype=tf.int32),
                tf.random.uniform((16, 8, 1), maxval=10, dtype=tf.int32),
            ]
        )
        layer = layers.RandomZoomAndCrop(
            height=8, width=8, scale_factor=(3 / 4, 4 / 3)
        )

        @tf.function
        def fn(inputs):
            outputs = layer(inputs)
            image_shape = outputs["images"].shape
            segmentation_mask_shape = outputs["segmentation_masks"].shape
            assert image_shape == (2, 8, 8, 3)
            assert segmentation_mask_shape == (2, 8, 8, 1)

        fn({"images": images, "segmentation_masks": segmentation_masks})



import tensorflow as tf

from keras_aug import layers


class RandomGridMaskTest(tf.test.TestCase):
    def test_one_channel(self):
        xs = tf.cast(
            tf.stack(
                [3 * tf.ones((80, 80, 1)), 2 * tf.ones((80, 80, 1))],
                axis=0,
            ),
            dtype=tf.float32,
        )
        fill_value = 0.0
        layer = layers.RandomGridMask(
            size_factor=(0.7, 1.0),
            ratio_factor=0.3,
            rotation_factor=(20, 30),
            fill_mode="constant",
            fill_value=fill_value,
            seed=2023,
        )

        xs = layer(xs)

        # Some pixels should be replaced with fill_value
        self.assertTrue(tf.math.reduce_any(xs[0] == float(fill_value)))
        self.assertTrue(tf.math.reduce_any(xs[0] == 3.0))
        self.assertTrue(tf.math.reduce_any(xs[1] == float(fill_value)))
        self.assertTrue(tf.math.reduce_any(xs[1] == 2.0))

    def test_non_square_image(self):
        xs = tf.cast(
            tf.stack(
                [2 * tf.ones((80, 40, 1)), tf.ones((80, 40, 1))],
                axis=0,
            ),
            dtype=tf.float32,
        )
        fill_value = 100.0
        layer = layers.RandomGridMask(
            size_factor=(0.5, 1.0),
            ratio_factor=0.6,
            rotation_factor=30,
            fill_mode="constant",
            fill_value=fill_value,
            seed=2023,
        )

        xs = layer(xs)

        # Some pixels should be replaced with fill_value
        self.assertTrue(tf.math.reduce_any(xs[0] == float(fill_value)))
        self.assertTrue(tf.math.reduce_any(xs[0] == 2.0))
        self.assertTrue(tf.math.reduce_any(xs[1] == float(fill_value)))
        self.assertTrue(tf.math.reduce_any(xs[1] == 1.0))

    def test_in_single_image(self):
        xs = tf.cast(
            tf.ones((8, 8, 1)),
            dtype=tf.float32,
        )
        layer = layers.RandomGridMask(
            ratio_factor=(0.5, 0.5), fill_mode="constant", fill_value=0.0
        )

        xs = layer(xs)

        self.assertTrue(tf.math.reduce_any(xs == 0.0))
        self.assertTrue(tf.math.reduce_any(xs == 1.0))



import inspect

import tensorflow as tf
from absl.testing import parameterized

from keras_aug import layers
from keras_aug.datapoints import bounding_box
from keras_aug.layers import augmentation
from keras_aug.layers import preprocessing
from keras_aug.utils.augmentation import BOUNDING_BOXES
from keras_aug.utils.augmentation import IMAGES
from keras_aug.utils.augmentation import LABELS

#   (
#       name,
#       layer_cls,
#       args,
#       is_bbox_compatible,
#   )
# all configurations should be expanded for readability
GENERAL_TESTS = [
    (
        "RandAugment",
        layers.RandAugment,
        {"value_range": (0, 255), "seed": 2023},
        True,
    ),
    (
        "TrivialAugmentWide",
        layers.TrivialAugmentWide,
        {"value_range": (0, 255)},
        True,
    ),
    (
        "RandomAffine",
        layers.RandomAffine,
        {
            "rotation_factor": 10,
            "translation_height_factor": 0.1,
            "translation_width_factor": 0.1,
            "zoom_height_factor": 0.1,
            "zoom_width_factor": 0.1,
            "shear_height_factor": 0.1,
            "shear_width_factor": 0.1,
        },
        True,
    ),
    (
        "RandomCrop",
        layers.RandomCrop,
        {"height": 2, "width": 2},
        True,
    ),
    (
        "RandomCropAndResize",
        layers.RandomCropAndResize,
        {
            "height": 2,
            "width": 2,
            "crop_area_factor": (0.8, 1.0),
            "aspect_ratio_factor": (3 / 4, 4 / 3),
        },
        True,
    ),
    (
        "RandomFlip",
        layers.RandomFlip,
        {"mode": "horizontal"},
        True,
    ),
    (
        "RandomResize",
        layers.RandomResize,
        {"heights": [2]},
        True,
    ),
    (
        "RandomRotate",
        layers.RandomRotate,
        {"factor": 10},
        True,
    ),
    (
        "RandomZoomAndCrop",
        layers.RandomZoomAndCrop,
        {"height": 2, "width": 2, "scale_factor": (0.8, 1.25)},
        True,
    ),
    (
        "MixUp",
        layers.MixUp,
        {},
        True,
    ),
    (
        "Mosaic",
        layers.Mosaic,
        {
            "height": 2,
            "width": 2,
        },
        True,
    ),
    (
        "ChannelShuffle",
        layers.ChannelShuffle,
        {"groups": 3},
        True,
    ),
    (
        "RandomBlur",
        layers.RandomBlur,
        {"factor": (3, 7)},
        True,
    ),
    (
        "RandomChannelShift",
        layers.RandomChannelShift,
        {"value_range": (0, 255), "factor": 0.1},
        True,
    ),
    (
        "RandomCLAHE",
        layers.RandomCLAHE,
        {"value_range": (0, 255)},
        True,
    ),
    (
        "RandomColorJitter",
        layers.RandomColorJitter,
        {
            "value_range": (0, 255),
            "brightness_factor": 0.1,
            "contrast_factor": 0.1,
            "saturation_factor": 0.1,
            "hue_factor": 0.1,
        },
        True,
    ),
    (
        "RandomGamma",
        layers.RandomGamma,
        {"value_range": (0, 255), "factor": 0.1},
        True,
    ),
    (
        "RandomGaussianBlur",
        layers.RandomGaussianBlur,
        {"kernel_size": 3, "factor": 2.0},
        True,
    ),
    (
        "RandomHSV",
        layers.RandomHSV,
        {
            "value_range": (0, 255),
            "hue_factor": 0.1,
            "saturation_factor": 0.1,
            "value_factor": 0.1,
        },
        True,
    ),
    (
        "RandomJpegQuality",
        layers.RandomJpegQuality,
        {
            "value_range": (0, 255),
            "factor": (75, 100),
        },
        True,
    ),
    (
        "RandomPosterize",
        layers.RandomPosterize,
        {"value_range": (0, 255), "factor": (5, 8)},
        True,
    ),
    (
        "RandomSharpness",
        layers.RandomSharpness,
        {"value_range": (0, 255), "factor": 0.1},
        True,
    ),
    (
        "RandomSolarize",
        layers.RandomSolarize,
        {
            "value_range": (0, 255),
            "threshold_factor": 10,
            "addition_factor": 10,
        },
        True,
    ),
    (
        "RandomChannelDropout",
        layers.RandomChannelDropout,
        {},
        True,
    ),
    (
        "RandomCutout",
        layers.RandomCutout,
        {"height_factor": 0.3, "width_factor": 0.3},
        True,
    ),
    (
        "RandomGridMask",
        layers.RandomGridMask,
        {
            "size_factor": (0.5, 1.0),
            "ratio_factor": (0.6, 0.6),
            "rotation_factor": (-10, 10),
        },
        True,
    ),
    (
        "RandomApply",
        layers.RandomApply,
        {"layer": layers.RandomChannelDropout()},
        True,
    ),
    (
        "RandomChoice",
        layers.RandomChoice,
        {
            "layers": [
                layers.RandomChannelDropout(),
                layers.RandomChannelDropout(),
            ]
        },
        True,
    ),
    (
        "RepeatedAugment",
        layers.RepeatedAugment,
        {
            "layers": [
                layers.RandomColorJitter(
                    value_range=(0, 255), brightness_factor=(1.5, 1.5)
                ),
                layers.RandomColorJitter(
                    value_range=(0, 255), contrast_factor=(1.5, 1.5)
                ),
            ]
        },
        True,
    ),
    (
        "CenterCrop",
        layers.CenterCrop,
        {"height": 2, "width": 2},
        True,
    ),
    (
        "PadIfNeeded",
        layers.PadIfNeeded,
        {"min_height": 2, "min_width": 2},
        True,
    ),
    (
        "Resize",
        layers.Resize,
        {"height": 2, "width": 2},
        True,
    ),
    (
        "AutoContrast",
        layers.AutoContrast,
        {"value_range": (0, 255)},
        True,
    ),
    (
        "Equalize",
        layers.Equalize,
        {"value_range": (0, 255)},
        True,
    ),
    (
        "Grayscale",
        layers.Grayscale,
        {"output_channels": 3},
        True,
    ),
    (
        "Identity",
        layers.Identity,
        {},
        True,
    ),
    (
        "Invert",
        layers.Invert,
        {"value_range": (0, 255)},
        True,
    ),
    (
        "Normalize",
        layers.Normalize,
        {"value_range": (0, 255)},
        True,
    ),
    (
        "Rescale",
        layers.Rescale,
        {"scale": 1.0 / 255.0},
        True,
    ),
    (
        "SanitizeBoundingBox",
        layers.SanitizeBoundingBox,
        {"min_size": 10},
        True,
    ),
]

NO_BOUNDING_BOXES_SUPPORT = [
    (
        "AugMix",
        layers.AugMix,
        {"value_range": (0, 255)},
        False,
    ),
    (
        "CutMix",
        layers.CutMix,
        {"alpha": 1.0},
        False,
    ),
    (
        "RandomErase",
        layers.RandomErase,
        {"area_factor": (0.02, 0.4), "aspect_ratio_factor": (0.3, 1.0 / 0.3)},
        False,
    ),
]


class WithBoundingBoxesTest(tf.test.TestCase, parameterized.TestCase):
    def test_all_2d_aug_layers_included(self):
        base_cls = layers.VectorizedBaseRandomLayer
        cls_spaces = [augmentation, preprocessing]
        all_2d_aug_layers = []
        for cls_space in cls_spaces:
            all_2d_aug_layers.extend(
                inspect.getmembers(cls_space, predicate=inspect.isclass)
            )
        all_2d_aug_layer_names = set(
            item[0]
            for item in all_2d_aug_layers
            if issubclass(item[1], base_cls)
        )

        cosistent_names = set(item[0] for item in GENERAL_TESTS)
        no_support_names = set(item[0] for item in NO_BOUNDING_BOXES_SUPPORT)
        all_test_names = cosistent_names.union(no_support_names)

        for name in all_2d_aug_layer_names:
            self.assertIn(name, all_test_names, msg=f"{name} not found")

    @parameterized.named_parameters(*GENERAL_TESTS)
    def test_ragged_bounding_boxes(self, layer_cls, args, is_bbox_compatible):
        images = tf.random.uniform(shape=(2, 16, 16, 3)) * 255.0
        labels = tf.ones(shape=(2, 1))
        bounding_boxes = {
            "boxes": tf.ragged.constant(
                [
                    [[1, 1, 2, 2], [10, 10, 15, 15]],
                    [[2, 2, 4, 4]],
                ],
                dtype=tf.float32,
            ),
            "classes": tf.ragged.constant([[0, 1], [2]], dtype=tf.float32),
        }
        if is_bbox_compatible:
            try:
                layer = layer_cls(**args, bounding_box_format="xyxy")
            except TypeError:
                layer = layer_cls(**args)
            inputs = {
                IMAGES: images,
                LABELS: labels,
                BOUNDING_BOXES: bounding_boxes,
            }
        else:
            layer = layer_cls(**args)
            inputs = {IMAGES: images, LABELS: labels}

        outputs = layer(inputs)

        bounding_boxes = outputs[BOUNDING_BOXES]
        bounding_box.validate_format(bounding_boxes)
        self.assertTrue(isinstance(bounding_boxes["boxes"], tf.RaggedTensor))

    @parameterized.named_parameters(*GENERAL_TESTS)
    def test_dense_bounding_boxes(self, layer_cls, args, is_bbox_compatible):
        images = tf.random.uniform(shape=(2, 16, 16, 3)) * 255.0
        labels = tf.random.uniform(shape=(2, 1)) * 10.0
        bounding_boxes = {
            "boxes": tf.ragged.constant(
                [
                    [[1, 1, 2, 2], [10, 10, 15, 15]],
                    [[2, 2, 4, 4]],
                ],
                dtype=tf.float32,
            ),
            "classes": tf.ragged.constant([[0, 1], [2]], dtype=tf.float32),
        }
        bounding_boxes = bounding_box.to_dense(bounding_boxes)
        if is_bbox_compatible:
            try:
                layer = layer_cls(**args, bounding_box_format="xyxy")
            except TypeError:
                layer = layer_cls(**args)
            inputs = {
                IMAGES: images,
                LABELS: labels,
                BOUNDING_BOXES: bounding_boxes,
            }
        else:
            layer = layer_cls(**args)
            inputs = {IMAGES: images, LABELS: labels}

        outputs = layer(inputs)

        bounding_boxes = outputs[BOUNDING_BOXES]
        bounding_box.validate_format(bounding_boxes)
        self.assertTrue(isinstance(bounding_boxes["boxes"], tf.Tensor))



from typing import Sequence

import tensorflow as tf
from tensorflow import keras

from keras_aug import core
from keras_aug.datapoints import image as image_utils
from keras_aug.layers.base.vectorized_base_random_layer import (
    VectorizedBaseRandomLayer,
)
from keras_aug.utils import augmentation as augmentation_utils


@keras.utils.register_keras_serializable(package="keras_aug")
class RandomJpegQuality(VectorizedBaseRandomLayer):
    """Randomly applies jpeg compression artifacts to the input images.

    Performs the jpeg compression algorithm on the image. This layer can be used
    in order to ensure your model is robust to artifacts introduced by JPEG
    compression.

    Args:
        value_range (Sequence[int|float]): The range of values the incoming
            images will have. This is typically either ``[0, 1]`` or
            ``[0, 255]`` depending on how your preprocessing pipeline is set up.
        factor (int|Sequence[int]|keras_aug.FactorSampler): The range of the
            compression factor. When represented as a single int, the
            factor will be randomly picked between ``[100 - factor, 100]``.
            ``50`` will give the image with 50% JPEG compression. ``100`` will
            still give a lossy compresson. This value is passed to
            ``tf.image.adjust_jpeg_quality()``.
        seed (int|float, optional): The random seed. Defaults to
            ``None``.

    References:
        - `KerasCV <https://github.com/keras-team/keras-cv>`_
    """

    def __init__(
        self,
        value_range,
        factor,
        seed=None,
        **kwargs,
    ):
        super().__init__(seed=seed, **kwargs)
        if isinstance(factor, (int, float)):
            lower = 101 - int(factor)
            upper = 101
            factor = (lower, upper)
        elif isinstance(factor, Sequence):
            factor = (factor[0], factor[1] + 1)
        elif isinstance(factor, core.FactorSampler):
            factor = factor
        else:
            raise ValueError(
                "RandomJpegQuality expects factor to be a list or a tuple of "
                f"ints. Got factor = {factor}"
            )
        self.factor = augmentation_utils.parse_factor(
            factor, min_value=0, max_value=100 + 1, seed=seed
        )
        self.value_range = value_range
        self.seed = seed

    def get_random_transformation_batch(self, batch_size, **kwargs):
        # scale from [0, 1] to [0, 100]
        factors = self.factor(shape=(batch_size, 1), dtype=tf.int32)
        return factors

    def augment_ragged_image(self, image, transformation, **kwargs):
        image = tf.expand_dims(image, axis=0)
        transformation = tf.expand_dims(transformation, axis=0)
        image = self.augment_images(
            images=image, transformations=transformation, **kwargs
        )
        return tf.squeeze(image, axis=0)

    def augment_images(self, images, transformations, **kwargs):
        images = image_utils.transform_value_range(
            images, self.value_range, target_range=(0, 1)
        )
        inputs_for_adjust_jpeg_qualitye = {
            "images": images,
            "factors": transformations,
        }
        images = tf.vectorized_map(
            self.adjust_jpeg_quality,
            inputs_for_adjust_jpeg_qualitye,
        )
        images = image_utils.transform_value_range(
            images, (0, 1), self.value_range, dtype=self.compute_dtype
        )
        return images

    def augment_labels(self, labels, transformations, **kwargs):
        return labels

    def augment_bounding_boxes(self, bounding_boxes, transformations, **kwargs):
        return bounding_boxes

    def augment_segmentation_masks(
        self, segmentation_masks, transformations, **kwargs
    ):
        return segmentation_masks

    def augment_keypoints(self, keypoints, transformations, **kwargs):
        return keypoints

    def adjust_jpeg_quality(self, inputs):
        image = inputs.get("images", None)

        image = tf.cast(image, tf.float32)
        factor = inputs.get("factors", None)

        image = tf.image.adjust_jpeg_quality(image, jpeg_quality=factor[0])
        return image

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "value_range": self.value_range,
                "factor": self.factor,
                "seed": self.seed,
            }
        )
        return config



from tensorflow import keras

from keras_aug.layers.base.vectorized_base_random_layer import (
    VectorizedBaseRandomLayer,
)


@keras.utils.register_keras_serializable(package="keras_aug")
class Identity(VectorizedBaseRandomLayer):
    """Applies nothing to the inputs."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def augment_ragged_image(self, image, transformation, **kwargs):
        return image

    def augment_images(self, images, transformations, **kwargs):
        return images

    def augment_labels(self, labels, transformations, **kwargs):
        return labels

    def augment_bounding_boxes(self, bounding_boxes, transformations, **kwargs):
        return bounding_boxes

    def augment_segmentation_masks(
        self, segmentation_masks, transformations, **kwargs
    ):
        return segmentation_masks

    def augment_keypoints(self, keypoints, transformations, **kwargs):
        return keypoints

    def get_config(self):
        return super().get_config()



import tensorflow as tf


def stateless_random_beta(
    shape, seed_alpha, seed_beta, alpha, beta, dtype=tf.float32
):
    seed_alpha = tf.cast(seed_alpha, dtype=tf.int32)
    seed_beta = tf.cast(seed_beta, dtype=tf.int32)
    sampled_alpha = tf.random.stateless_gamma(
        shape, seed_alpha, alpha=alpha, dtype=dtype
    )
    sampled_beta = tf.random.stateless_gamma(
        shape, seed_beta, alpha=beta, dtype=dtype
    )
    return sampled_alpha / (sampled_alpha + sampled_beta)


def stateless_random_dirichlet(shape, seed, alpha, dtype=tf.float32):
    sampled_gamma = tf.random.stateless_gamma(
        shape, seed, alpha=alpha, dtype=dtype
    )
    return sampled_gamma / tf.reduce_sum(sampled_gamma, axis=-1, keepdims=True)



"""
Most of these codes come from KerasCV.
"""
import numpy as np
import tensorflow as tf

from keras_aug.datapoints import bounding_box


class IoUTest(tf.test.TestCase):
    def test_compute_single_iou(self):
        bb1 = tf.constant([[100, 101, 200, 201]], dtype=tf.float32)
        bb1_off_by_1 = tf.constant([[101, 102, 201, 202]], dtype=tf.float32)
        # area of bb1 and bb1_off_by_1 are each 10000.
        # intersection area is 99*99=9801
        # iou=9801/(2*10000 - 9801)=0.96097656633
        self.assertAlmostEqual(
            bounding_box.compute_iou(bb1, bb1_off_by_1, "yxyx")[0],
            0.96097656633,
        )

    def test_compute_iou(self):
        bb1 = [100, 101, 200, 201]
        bb1_off_by_1_pred = [101, 102, 201, 202]
        iou_bb1_bb1_off = 0.96097656633
        top_left_bounding_box = [0, 2, 1, 3]
        far_away_box = [1300, 1400, 1500, 1401]
        another_far_away_pred = [1000, 1400, 1200, 1401]

        # Rows represent predictions, columns ground truths
        expected_result = np.array(
            [[iou_bb1_bb1_off, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]],
            dtype=np.float32,
        )

        sample_y_true = tf.constant(
            [bb1, top_left_bounding_box, far_away_box], dtype=tf.float32
        )
        sample_y_pred = tf.constant(
            [bb1_off_by_1_pred, top_left_bounding_box, another_far_away_pred],
            dtype=tf.float32,
        )

        result = bounding_box.compute_iou(sample_y_true, sample_y_pred, "yxyx")
        self.assertAllClose(expected_result, result.numpy())

    def test_batched_compute_iou(self):
        bb1 = [100, 101, 200, 201]
        bb1_off_by_1_pred = [101, 102, 201, 202]
        iou_bb1_bb1_off = 0.96097656633
        top_left_bounding_box = [0, 2, 1, 3]
        far_away_box = [1300, 1400, 1500, 1401]
        another_far_away_pred = [1000, 1400, 1200, 1401]

        # Rows represent predictions, columns ground truths
        expected_result = np.array(
            [
                [[iou_bb1_bb1_off, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]],
                [[iou_bb1_bb1_off, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]],
            ],
            dtype=np.float32,
        )

        sample_y_true = tf.constant(
            [
                [bb1, top_left_bounding_box, far_away_box],
                [bb1, top_left_bounding_box, far_away_box],
            ],
            dtype=tf.float32,
        )
        sample_y_pred = tf.constant(
            [
                [
                    bb1_off_by_1_pred,
                    top_left_bounding_box,
                    another_far_away_pred,
                ],
                [
                    bb1_off_by_1_pred,
                    top_left_bounding_box,
                    another_far_away_pred,
                ],
            ],
            dtype=tf.float32,
        )

        result = bounding_box.compute_iou(sample_y_true, sample_y_pred, "yxyx")
        self.assertAllClose(expected_result, result.numpy())

    def test_batched_boxes1_unbatched_boxes2(self):
        bb1 = [100, 101, 200, 201]
        bb1_off_by_1_pred = [101, 102, 201, 202]
        iou_bb1_bb1_off = 0.96097656633
        top_left_bounding_box = [0, 2, 1, 3]
        far_away_box = [1300, 1400, 1500, 1401]
        another_far_away_pred = [1000, 1400, 1200, 1401]

        # Rows represent predictions, columns ground truths
        expected_result = np.array(
            [
                [[iou_bb1_bb1_off, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]],
                [[iou_bb1_bb1_off, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]],
            ],
            dtype=np.float32,
        )

        sample_y_true = tf.constant(
            [
                [bb1, top_left_bounding_box, far_away_box],
                [bb1, top_left_bounding_box, far_away_box],
            ],
            dtype=tf.float32,
        )
        sample_y_pred = tf.constant(
            [bb1_off_by_1_pred, top_left_bounding_box, another_far_away_pred],
            dtype=tf.float32,
        )

        result = bounding_box.compute_iou(sample_y_true, sample_y_pred, "yxyx")
        self.assertAllClose(expected_result, result.numpy())

    def test_unbatched_boxes1_batched_boxes2(self):
        bb1 = [100, 101, 200, 201]
        bb1_off_by_1_pred = [101, 102, 201, 202]
        iou_bb1_bb1_off = 0.96097656633
        top_left_bounding_box = [0, 2, 1, 3]
        far_away_box = [1300, 1400, 1500, 1401]
        another_far_away_pred = [1000, 1400, 1200, 1401]

        # Rows represent predictions, columns ground truths
        expected_result = np.array(
            [
                [[iou_bb1_bb1_off, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]],
                [[iou_bb1_bb1_off, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]],
            ],
            dtype=np.float32,
        )

        sample_y_true = tf.constant(
            [
                [bb1, top_left_bounding_box, far_away_box],
            ],
            dtype=tf.float32,
        )
        sample_y_pred = tf.constant(
            [
                [
                    bb1_off_by_1_pred,
                    top_left_bounding_box,
                    another_far_away_pred,
                ],
                [
                    bb1_off_by_1_pred,
                    top_left_bounding_box,
                    another_far_away_pred,
                ],
            ],
            dtype=tf.float32,
        )

        result = bounding_box.compute_iou(sample_y_true, sample_y_pred, "yxyx")
        self.assertAllClose(expected_result, result.numpy())



import tensorflow as tf
from tensorflow import keras


@keras.utils.register_keras_serializable(package="keras_aug")
class FactorSampler:
    """FactorSampler represents a strength factor for use in an augmentation
    layer.

    FactorSampler should be subclassed and implement a `__call__()` method that
    returns a tf.float32, or a float. This method will be used by preprocessing
    layers to determine the strength of their augmentation. The specific range
    of values supported may vary by layer, but for most layers is the range
    [0, 1].

    References:
        - `KerasCV <https://github.com/keras-team/keras-cv>`_
    """

    def __call__(self, shape=None, dtype=tf.float32):
        raise NotImplementedError(
            "FactorSampler subclasses must implement a `__call__()` method."
        )

    def get_config(self):
        return {}

    @classmethod
    def from_config(cls, config):
        return cls(**config)



import tensorflow as tf
from tensorflow import keras

from keras_aug.layers.base.vectorized_base_random_layer import (
    VectorizedBaseRandomLayer,
)


@keras.utils.register_keras_serializable(package="keras_aug")
class Normalize(VectorizedBaseRandomLayer):
    """Normalizes the mean and std on given images.

    Normalize applies following equation to the input images:
    ``y = (x - mean * max_pixel_value) / (std * max_pixel_value)``

    Args:
        value_range (Sequence[int|float]): The range of values the incoming
            images will have. This is typically either ``[0, 1]`` or
            ``[0, 255]`` depending on how your preprocessing pipeline is set up.
        mean (list(float)): The mean values. Defaults to
            ``(0.485, 0.456, 0.406)`` which is the mean values from ImageNet.
        std (list(float)): The std values. Defaults to
            ``(0.229, 0.224, 0.225)`` which is the std values from ImageNet
        seed (int|float, optional): The random seed. Defaults to
            ``None``.
    """

    def __init__(
        self,
        value_range,
        mean=(0.485, 0.456, 0.406),
        std=(0.229, 0.224, 0.225),
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.value_range = value_range
        self.mean_input = mean
        self.std_input = std

        num_channel = len(mean)
        self.mean = tf.convert_to_tensor(mean, dtype=self.compute_dtype)
        self.mean = tf.reshape(self.mean, shape=(1, 1, 1, num_channel))
        self.std = tf.convert_to_tensor(std, dtype=self.compute_dtype)
        self.std = tf.reshape(self.std, shape=(1, 1, 1, num_channel))

    def augment_ragged_image(self, image, transformation, **kwargs):
        images = tf.expand_dims(image, axis=0)
        images = self.augment_images(
            images=images, transformations=transformation, **kwargs
        )
        return tf.squeeze(images, axis=0)

    def augment_images(self, images, transformations, **kwargs):
        max_pixel_value = self.value_range[1]
        images = tf.cast(images, dtype=self.compute_dtype)
        images = (images - self.mean * max_pixel_value) / (
            self.std * max_pixel_value
        )
        return images

    def augment_labels(self, labels, transformations, **kwargs):
        return labels

    def augment_bounding_boxes(self, bounding_boxes, transformations, **kwargs):
        return bounding_boxes

    def augment_segmentation_masks(
        self, segmentation_masks, transformations, **kwargs
    ):
        return segmentation_masks

    def augment_keypoints(self, keypoints, transformations, **kwargs):
        return keypoints

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "value_range": self.value_range,
                "mean": self.mean_input,
                "std": self.std_input,
            }
        )
        return config
